import os
from pathlib import Path
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field

from laktory import settings
from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

logger = get_logger(__name__)


class PythonPackage(BaseModel, PulumiResource, TerraformResource):
    """
    Python Package built and deployed as a wheel file.

    This resource type allows to target a local python package to generate a wheel file
    that will be built at deploy and deployed as a WorkspaceFile.

    Examples
    --------
    ```py
    import laktory as lk

    pp = lk.models.resources.databricks.PythonPackage(
        package_name="lake",
        config_filepath="lake/pyproject.toml",
        dirpath="/wheels/",
    )
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    build_command: str = Field(
        "uv build --wheel",
        description="The build command used to generate the wheel file.",
    )
    config_filepath: str = Field(
        ...,
        description="File path of the pyproject.toml file or setup.py configuration file.",
    )
    dirpath: str = Field(
        None,
        description="Workspace directory inside rootpath in which the workspace file is deployed. Used only if `path` is not specified.",
    )
    package_name: str = Field(..., description="Name of the package")
    path_: str = Field(
        None,
        description="Workspace filepath for the file. Overwrite `rootpath` and `dirpath`.",
        validation_alias=AliasChoices("path_", "path"),
        exclude=True,
    )
    _wheel_path: Path = None

    @classmethod
    def lookup_defaults(cls) -> dict:
        return {"path": ""}

    @property
    def filename(self) -> str | None:
        """File filename"""
        if self.source:
            return os.path.basename(self.source)

    def get_path(self):
        """Get workspace path"""
        # Path set
        if self.path_:
            return self.path_

        if not self.source:
            return None

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        # path
        _path = Path(settings.workspace_laktory_root) / self.dirpath / self.filename

        return _path.as_posix()

    @computed_field(description="path")
    @property
    def path(self) -> str | None:
        return self.get_path()

    @computed_field(description="Wheel file path", return_type=str)
    @property
    def source(self) -> str:
        return str(self._wheel_path)

    def build(self):
        # Ideally we would want to deterministically find the value of wheel_path
        # without building it, but it seems cumbersome and unreliable to parse
        # the config file and find the version.
        if self._wheel_path is None:
            from laktory.cli._common import Worker

            if not Path(self.config_filepath).exists():
                raise ValueError(
                    f"Config filepath {Path(self.config_filepath).absolute().resolve()} ({self.config_filepath}) for package '{self.package_name}' doest not exists"
                )

            package_root = Path(self.config_filepath).parent
            dist_path = package_root / "dist"

            worker = Worker()

            cmd = self.build_command.split(" ")

            logger.info(
                f"Building package '{self.package_name}' from {package_root} with '{' '.join(cmd)}'"
            )
            worker.run(
                cmd=cmd,
                cwd=package_root,
                raise_exceptions=settings.cli_raise_external_exceptions,
            )

            files = [f for f in dist_path.glob("*.whl") if f.is_file()]
            if len(files) == 0:
                raise RuntimeError(
                    "Wheel file could not be built. Make sure the `config_filepath` and `build_command` are correct."
                )
            file = max(files, key=lambda f: f.stat().st_mtime)

            self._wheel_path = file

            # Renaming wheel file
            # if self.wheel_filename:
            #     new_path = self._wheel_path.with_name(self.wheel_filename)
            #     logger.info(f"Renaming {self._wheel_path} -> {new_path}")
            #     self._wheel_path.rename(new_path)
            #     self._wheel_path = new_path

            self._wheel_path = self._wheel_path.absolute().resolve()

        return self._wheel_path

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """package name"""
        return self.package_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []
        if self.access_controls:
            from laktory.models.resources.databricks.pythonpackagepermissions import (
                PythonPackagePermissions,
            )

            resources += [
                PythonPackagePermissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    get_workspace_file_path=self.get_path,
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:WorkspaceFile"

    @property
    def pulumi_excludes(self) -> list[str] | dict[str, bool]:
        return [
            "access_controls",
            "build_command",
            "config_filepath",
            "dirpath",
            "package_name",
            "wheel_filename",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_workspace_file"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import laktory.models.resources.databricks
import laktory.models.resources.providers
from laktory.models.resources.baseresource import BaseResource
from laktory.models.resources.providers import *
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource



from typing import Annotated

import typer

from laktory.cli._common import CLIController
from laktory.cli.app import app
from laktory.constants import SUPPORTED_BACKENDS


@app.command()
def preview(
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
    options: Annotated[
        str,
        typer.Option("--options", help="Comma separated IaC backend options (flags)."),
    ] = None,
):
    """
    Validate configuration and resources and preview deployment.

    Parameters
    ----------
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.
    options:
        Comma separated IaC backend options (flags).

    Examples
    --------
    ```cmd
    laktory preview --env dev pulumi_options "--show-reads,--show-config"
    ```

    References
    ----------
    * [CLI](https://www.laktory.ai/concepts/cli/)
    * pulumi [preview](https://www.pulumi.com/docs/cli/commands/pulumi_preview/)
    * terraform [preview](https://developer.hashicorp.com/terraform/cli/commands/plan)
    """
    controller = CLIController(
        env=environment,
        stack_filepath=filepath,
        options_str=options,
    )

    # Call
    if controller.backend == "pulumi":
        controller.pulumi_call("preview")
    elif controller.backend == "terraform":
        controller.terraform_call("plan")
    else:
        raise ValueError(f"backend should be {SUPPORTED_BACKENDS}")



from typing import Literal
from typing import Union

from pydantic import AliasChoices
from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ClusterAutoScale(BaseModel):
    min_workers: int = Field(..., description="Minimum number of worker nodes")
    max_workers: int = Field(..., description="Maximum number of worker nodes")


class ClusterInitScriptVolumes(BaseModel):
    destination: str = Field(None, description="Volume filepath")


class ClusterInitScriptWorkspace(BaseModel):
    destination: str = Field(None, description="Workspace filepath")


class ClusterInitScript(BaseModel):
    volumes: ClusterInitScriptVolumes = Field(
        None, description="Volumes file specification"
    )
    workspace: ClusterInitScriptWorkspace = Field(
        None, description="Workspace file specifications"
    )


class ClusterLibraryCran(BaseModel):
    package: str = Field(None, description="")
    repo: str = Field(None, description="")


class ClusterLibraryMaven(BaseModel):
    coordinates: str = Field(None, description="")
    exclusions: list[str] = Field(None, description="")
    repo: str = Field(None, description="")


class ClusterLibraryPypi(BaseModel):
    package: str = Field(None, description="Package name")
    repo: str = Field(None, description="Packages repository")


class ClusterLibrary(BaseModel):
    cran: ClusterLibraryCran = Field(None, description="Cran library specifications")
    egg: str = Field(None, description="Egg filepath")
    jar: str = Field(None, description="Jar filepath")
    maven: ClusterLibraryMaven = Field(None, description="Maven specifications")
    pypi: ClusterLibraryPypi = Field(None, description="Pypi library specifications")
    whl: str = Field(None, description="Wheel filepath")


class ClusterLookup(ResourceLookup):
    cluster_id: str = Field(
        serialization_alias="id", description="The id of the cluster"
    )


class Cluster(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks cluster

    Examples
    --------
    ```py
    from laktory import models

    cluster = models.resources.databricks.Cluster(
        name="default",
        spark_version="16.3.x-scala2.12",
        data_security_mode="USER_ISOLATION",
        node_type_id="Standard_DS3_v2",
        autoscale={
            "min_workers": 1,
            "max_workers": 4,
        },
        num_workers=0,
        autotermination_minutes=30,
        libraries=[{"pypi": {"package": "laktory==0.0.23"}}],
        access_controls=[
            {
                "group_name": "role-engineers",
                "permission_level": "CAN_RESTART",
            }
        ],
        is_pinned=True,
    )
    ```

    References
    ----------

    * [Databricks Cluster](https://docs.databricks.com/en/compute/configure.html#autoscaling-local-storage-1)
    * [Pulumi Databricks Cluster](https://www.pulumi.com/registry/packages/databricks/api-docs/cluster/)

    """

    access_controls: list[AccessControl] = Field(
        [], description="List of access controls"
    )
    apply_policy_default_values: bool = Field(
        None,
        description="Whether to use policy default values for missing cluster attributes.",
    )
    autoscale: ClusterAutoScale = Field(None, description="Autoscale specifications")
    autotermination_minutes: int = Field(
        None,
        description="Automatically terminate the cluster after being inactive for this time in minutes.",
    )
    # aws_attributes
    # azure_attributes
    cluster_id: str = Field(
        None, description="Cluster ID. Used when assigning a cluster to a job task."
    )
    cluster_name: str = Field(
        None,
        description="""
    Cluster name, which doesn’t have to be unique. If not specified at creation, the cluster name will be an empty string.
    """,
        validation_alias=AliasChoices("cluster_name", "name"),
    )

    # cluster_log_conf
    # cluster_source
    # cluster_mount_infos
    custom_tags: dict[str, str] = Field(
        None,
        description="""
    Additional tags for cluster resources. Databricks will tag all cluster resources (e.g., AWS EC2 instances and EBS
    volumes) with these tags in addition to default_tags. If a custom cluster tag has the same name as a default
    cluster tag, the custom tag is prefixed with an x_ when it is propagated.
    """,
    )
    data_security_mode: str = Field(
        None,
        description="""
    Select the security features of the cluster (see API docs for full list of values). Unity Catalog requires 
    SINGLE_USER or USER_ISOLATION mode. LEGACY_PASSTHROUGH for passthrough cluster and LEGACY_TABLE_ACL for Table ACL
    cluster. If omitted, default security features are enabled. To disable security features use NONE or legacy mode
    NO_ISOLATION. If kind is specified, then the following options are available:
    - DATA_SECURITY_MODE_AUTO: Databricks will choose the most appropriate access mode depending on your compute configuration.
    - DATA_SECURITY_MODE_STANDARD: Alias for USER_ISOLATION.
    - DATA_SECURITY_MODE_DEDICATED: Alias for SINGLE_USER.
        """,
    )
    # docker_image
    driver_instance_pool_id: str = Field(
        None,
        description="""
    Similar to instance_pool_id, but for driver node. If omitted, and instance_pool_id is specified, then the driver 
    will be allocated from that pool.
    """,
    )
    driver_node_type_id: str = Field(
        None,
        description="""
    The node type of the Spark driver. This field is optional; if unset, API will set the driver node type to the same
    value as node_type_id defined above.
    """,
    )
    enable_elastic_disk: bool = Field(
        None,
        description="""
    If you don’t want to allocate a fixed number of EBS volumes at cluster creation time, use autoscaling local 
    storage. With autoscaling local storage, Databricks monitors the amount of free disk space available on your
    cluster’s Spark workers. If a worker begins to run too low on disk, Databricks automatically attaches a new EBS 
    volume to the worker before it runs out of disk space. EBS volumes are attached up to a limit of 5 TB of total 
    disk space per instance (including the instance’s local storage). To scale down EBS usage, make sure you have
    autotermination_minutes and autoscale attributes set.
    """,
    )
    enable_local_disk_encryption: bool = Field(
        None,
        description="""
    Some instance types you use to run clusters may have locally attached disks. Databricks may store shuffle data or
    temporary data on these locally attached disks. To ensure that all data at rest is encrypted for all storage types,
    including shuffle data stored temporarily on your cluster’s local disks, you can enable local disk encryption. When
    local disk encryption is enabled, Databricks generates an encryption key locally unique to each cluster node and 
    uses it to encrypt all data stored on local disks. The scope of the key is local to each cluster node and is
    destroyed along with the cluster node itself. During its lifetime, the key resides in memory for encryption and 
    decryption and is stored encrypted on the disk. Your workloads may run more slowly because of the performance 
    impact of reading and writing encrypted data to and from local volumes. This feature is not available for all 
    Azure Databricks subscriptions. Contact your Microsoft or Databricks account representative to request access.
    """,
    )
    # gcp_attributes
    idempotency_token: str = Field(
        None,
        description="""
    An optional token to guarantee the idempotency of cluster creation requests. If an active cluster with the provided
    token already exists, the request will not create a new cluster, but it will return the existing running cluster's
    ID instead. If you specify the idempotency token, upon failure, you can retry until the request succeeds.
    Databricks platform guarantees to launch exactly one cluster with that idempotency token. This token should have 
    at most 64 characters.
    """,
    )
    init_scripts: list[ClusterInitScript] = Field(
        [], description="List of init scripts specifications"
    )
    instance_pool_id: str = Field(
        None,
        description="""
    To reduce cluster start time, you can attach a cluster to a predefined pool of idle instances. When attached to a 
    pool, a cluster allocates its driver and worker nodes from the pool. If the pool does not have sufficient idle 
    resources to accommodate the cluster’s request, it expands by allocating new instances from the instance provider. 
    When an attached cluster changes its state to TERMINATED, the instances it used are returned to the pool and 
    reused by a different cluster.
    """,
    )
    is_pinned: bool = Field(
        True,
        description="""
    boolean value specifying if the cluster is pinned (not pinned by default). You must be a Databricks administrator 
    to use this. The pinned clusters' maximum number is limited to 100, so apply may fail if you have more than that 
    (this number may change over time, so check Databricks documentation for actual number).
    """,
    )
    is_single_node: bool = Field(
        None,
        description="""
    When set to true, Databricks will automatically set single node related custom_tags, spark_conf, and num_workers.
        """,
    )
    kind: str = Field(
        None,
        description="""
    The kind of compute described by this compute specification. Possible values (see API docs for full list): 
    CLASSIC_PREVIEW (if corresponding public preview is enabled).
        """,
    )
    libraries: list[ClusterLibrary] = Field(
        [], description="List of libraries specifications"
    )
    lookup_existing: ClusterLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    node_type_id: str = Field(
        ...,
        description="Any supported databricks.getNodeType id. If instance_pool_id is specified, this field is not needed.",
    )
    no_wait: bool = Field(
        None,
        description="""
    If true, the provider will not wait for the cluster to reach RUNNING state when creating the cluster, allowing 
    cluster creation and library installation to continue asynchronously. Defaults to false (the provider will wait 
    for cluster creation and library installation to succeed).
    """,
    )
    num_workers: int = Field(
        None,
        description="""
    Number of worker nodes that this cluster should have. A cluster has one Spark driver and num_workers executors
    for a total of num_workers + 1 Spark nodes.
    """,
    )
    policy_id: str = Field(None, description="")
    runtime_engine: Literal["STANDARD", "PHOTON"] = Field(
        None,
        description="""
    The type of runtime engine to use. If not specified, the runtime engine type is inferred based on the 
    spark_version value
    """,
    )
    remote_disk_throughput: int = Field(None, description="""""")
    single_user_name: str = Field(
        None,
        description="""
    The optional user name of the user to assign to an interactive cluster. This field is required when using 
    data_security_mode set to SINGLE_USER or AAD Passthrough for Azure Data Lake Storage (ADLS) with a single-user
    cluster (i.e., not high-concurrency clusters).
    """,
    )
    spark_conf: dict[str, str] = Field(
        {},
        description="""
    Map with key-value pairs to fine-tune Spark clusters, where you can provide custom Spark configuration
    properties in a cluster configuration.
    """,
    )
    spark_env_vars: dict[str, str] = Field(
        {},
        description="""
    Map with environment variable key-value pairs to fine-tune Spark clusters. Key-value pairs of the form (X,Y)
    are exported (i.e., X='Y') while launching the driver and workers.
    """,
    )
    spark_version: str = Field(
        ...,
        description="""
    Runtime version of the cluster. Any supported databricks.getSparkVersion id. We advise using Cluster Policies
    to restrict the list of versions for simplicity while maintaining enough control.
    """,
    )
    ssh_public_keys: list[str] = Field(
        [],
        description="""
    SSH public key contents that will be added to each Spark node in this cluster. The corresponding private keys
    can be used to login with the user name ubuntu on port 2200. You can specify up to 10 keys.
    """,
    )
    total_initial_remote_disk_size: int = Field(None, description="""""")
    use_ml_runtime: bool = Field(
        None,
        description="""
    Whenever ML runtime should be selected or not. Actual runtime is determined by spark_version (DBR release), this
    field use_ml_runtime, and whether node_type_id is GPU node or not.
        """,
    )
    # workload_type:

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    cluster_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Cluster"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "init_scripts": "init_scripts",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_cluster"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import narwhals as nw
from planck import units


def convert_units(
    self,
    input_unit: str,
    output_unit: str,
) -> nw.Expr:
    """
    Units conversion

    Parameters
    ----------
    input_unit:
        Input units
    output_unit:
        Output units

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk  # noqa: F401

    df = nw.from_native(pl.DataFrame({"x": [1]}))
    df = df.with_columns(
        y=nw.col("x").laktory.convert_units(input_unit="m", output_unit="ft")
    )
    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    | | x | y       |  |
    | |---|---------|  |
    | | 1 | 3.28084 |  |
    └──────────────────┘
    '''
    ```

    References
    ----------
    The units conversion function use [planck](https://www.okube.ai/planck/) convert as a backend.
    """
    return units.convert(self._expr, input_unit, output_unit)



import os
from pathlib import Path
from typing import Literal
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field

from laktory import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class NotebookLookup(ResourceLookup):
    path: str = Field(
        serialization_alias="id", description="Notebook path on the workspace"
    )
    format: str = Field(
        "SOURCE",
        description="Notebook format to export. Either `SOURCE`, `HTML`, `JUPYTER`, or `DBC`",
    )


class Notebook(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Notebook

    Examples
    --------
    ```py
    from laktory import models

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
    )
    print(notebook.path)
    # > /.laktory/dlt_laktory_pl.py

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
    )
    print(notebook.path)
    # > /.laktory/dlt_laktory_pl.py

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
        dirpath="notebooks/dlt/",
    )
    print(notebook.path)
    # > /.laktory/notebooks/dlt/dlt_laktory_pl.py
    ```
    """

    access_controls: list[AccessControl] = Field(
        [], description="List of notebook access controls"
    )
    dirpath: str = Field(
        None,
        description="Workspace directory inside rootpath in which the notebook is deployed. Used only if `path` is not specified.",
    )
    language: Literal["SCALA", "PYTHON", "SQL", "R"] = Field(
        None, description="Notebook programming language"
    )
    lookup_existing: NotebookLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    path_: str = Field(
        None,
        description="Workspace filepath for the notebook. Overwrite `rootpath` and `dirpath`.",
        validation_alias=AliasChoices("path", "path_"),
        exclude=True,
    )
    source: str = Field(
        ..., description="Path to notebook in source code format on local filesystem."
    )

    @computed_field(description="path")
    @property
    def path(self) -> str | None:
        if self.path_:
            return self.path_

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        path = Path(settings.workspace_laktory_root) / self.dirpath / self.filename
        return path.as_posix()

    @property
    def filename(self) -> str:
        """Notebook file name"""
        return os.path.basename(self.source)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.path

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    notebook_path=f"${{resources.{self.resource_name}.path}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Notebook"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_notebook"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from .basedatasink import BaseDataSink
from .filedatasink import FileDataSink
from .hivemetastoredatasink import HiveMetastoreDataSink
from .mergecdcoptions import DataSinkMergeCDCOptions
from .pipelineviewdatasink import PipelineViewDataSink
from .tabledatasink import TableDataSink
from .tabledatasinkmetadata import TableDataSinkMetadata
from .unitycatalogdatasink import UnityCatalogDataSink

classes = [
    PipelineViewDataSink,
    FileDataSink,
    UnityCatalogDataSink,
    HiveMetastoreDataSink,
]


DataSinksUnion = (
    PipelineViewDataSink | FileDataSink | UnityCatalogDataSink | HiveMetastoreDataSink
)



from typing import Union

from pydantic import AliasChoices
from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Grant(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Grant

    Authoritative for a specific principal. Updates the grants of a securable to a
    single principal. Other principals within the grants for the securables are preserved.

    Examples
    --------
    ```py
    from laktory import models

    grants = models.resources.databricks.Grant(
        catalog="dev",
        principal="metastore-admins",
        privileges=["CREATE_SCHEMA"],
    )
    ```
    """

    catalog: str = Field(
        None, description="Name of the catalog to assign the grants to"
    )
    external_location: str = Field(
        None, description="Name of the external location to assign the grants to"
    )
    metastore: str = Field(
        None, description="Name of the metastore to assign the grants to"
    )
    model: str = Field(
        None, description="Name of the user to assign the permission to."
    )
    principal: str = Field(..., description="User, group or service principal name")
    privileges: list[str] = Field(..., description="List of allowed privileges")
    schema_: str = Field(
        None,
        validation_alias=AliasChoices("schema", "schema_"),
        description="Name of the schema to assign the permission to.",
    )  # required not to overwrite BaseModel attribute
    share: str = Field(
        None, description="Name of the share to assign the permission to."
    )
    storage_credential: str = Field(
        None, description="Name of the storage credential to assign the permission to."
    )
    table: str = Field(
        None, description="Name of the table to assign the permission to."
    )
    view: str = Field(None, description="Name of the view to assign the permission to.")
    volume: str = Field(
        None, description="Name of the volume to assign the permission to."
    )

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #
    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"schema_": "schema"}

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Grant"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return []

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_grant"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import narwhals as nw

from laktory._logger import get_logger
from laktory.models.dataframe.dataframecolumnexpr import DataFrameColumnExpr
from laktory.typing import AnyFrame

logger = get_logger(__name__)


def groupby_and_agg(
    self,
    groupby_columns: list[str] = None,
    agg_expressions: list[DataFrameColumnExpr | str | nw.Expr] = None,
) -> AnyFrame:
    """
    Apply a groupby and create aggregation columns.

    Parameters
    ----------
    groupby_columns:
        List of column names to group by
    agg_expressions:
        List of columns defining the aggregations

    Examples
    --------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk  # noqa: F401

    df0 = nw.from_native(
        pl.DataFrame(
            {
                "x": [0, 0, 1, 1],
                "y": [1, 2, 3, 4],
            }
        )
    )

    df = df0.laktory.groupby_and_agg(
        groupby_columns=["x"],
        agg_expressions=[
            "nw.col('y').mean().alias('mean_price')",
        ],
    )

    print(df.sort("x"))
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    || x | mean_price ||
    ||---|------------||
    || 0 | 1.5        ||
    || 1 | 3.5        ||
    └──────────────────┘
    '''
    ```
    """
    from laktory.models.dataframe.dataframecolumnexpr import DataFrameColumnExpr

    # Parse inputs
    if agg_expressions is None:
        raise ValueError("`agg_expressions` must be specified")
    if groupby_columns is None:
        groupby_columns = []

    logger.info(f"Executing groupby ({groupby_columns}) with {agg_expressions}")

    # Groupby arguments
    groupby = []

    for c in groupby_columns:
        groupby += [c]

    # Agg arguments
    aggs = []
    for expr in agg_expressions:
        if isinstance(expr, str):
            expr = DataFrameColumnExpr(expr=expr).to_expr()

        elif isinstance(expr, dict):
            expr = DataFrameColumnExpr(**expr).to_expr()

        aggs += [expr]

    return self._df.group_by(groupby).agg(*aggs)



from laktory._logger import get_logger
from laktory.enums import STREAMING_BACKENDS
from laktory.enums import DataFrameBackends
from laktory.models.dataframe.dataframecolumnexpr import DataFrameColumnExpr

# from laktory.spark.dataframe.watermark import Watermark
from laktory.typing import AnyFrame

logger = get_logger(__name__)


def stream_join(
    self,
    other: AnyFrame,
    how: str = "left",
    on: str | DataFrameColumnExpr | list[str | DataFrameColumnExpr] = None,
    left_on: str | DataFrameColumnExpr | list[str | DataFrameColumnExpr] = None,
    # left_watermark: Watermark = None,
    other_on: str | DataFrameColumnExpr | list[str | DataFrameColumnExpr] = None,
    # other_watermark: Watermark = None,
    time_constraint_interval_lower: str = "60 seconds",
    time_constraint_interval_upper: str = None,
    # coalesce: bool = False,
) -> AnyFrame:
    """
    Perform static-stream and stream-stream by applying watermarks and setting
    time constraints.

    Parameters
    ----------
    other:
        Right side of the join
    how:
        Type of join (left, outer, full, etc.)
    on:
        A list of strings for the columns to join on. The columns must exist
        on both sides.
    left_on:
        Name(s) of the left join column(s).
    left_watermark
        Watermark for left dataframe
    other_on:
        Name(s) of the right join column(s).
    other_watermark
        Watermark for other dataframe
    time_constraint_interval_lower:
        Lower bound for a spark streaming event-time constraint
    time_constraint_interval_upper:
        Upper bound for a spark streaming event-time constraint
    coalesce:
        If `True` columns present in both left and other are coalesced.

    Examples
    --------
    # ```py
    # import pandas as pd
    #
    # import laktory as lk  # noqa: F401
    #
    # df_prices = spark.createDataFrame(
    #     pd.DataFrame(
    #         {
    #             "symbol": ["AAPL", "GOOGL"],
    #             "price": [200.0, 205.0],
    #         }
    #     )
    # )
    #
    # df_meta = spark.createDataFrame(
    #     pd.DataFrame(
    #         {
    #             "symbol": ["AAPL", "GOOGL"],
    #             "name": ["Apple", "Google"],
    #         }
    #     )
    # )
    #
    # df = df_prices.laktory.smart_join(
    #     other=df_meta,
    #     on=["symbol"],
    # )
    #
    # print(df.toPandas().to_string())
    # '''
    #    price    name symbol
    # 0  200.0   Apple   AAPL
    # 1  205.0  Google  GOOGL
    # '''
    # ```

    References
    ----------

    * [pyspark join](https://spark.apache.org/docs/3.1.2/api/python/reference/api/pyspark.sql.DataFrame.join.html)
    * [spark streaming join](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#inner-joins-with-optional-watermarking)
    """

    left = self._df

    backend = DataFrameBackends.from_df(left)
    if backend not in STREAMING_BACKENDS:
        raise ValueError(f"Backend '{backend}' is not supported by stream_join.")

    logger.info(f"Executing {left} {how} JOIN {other}")

    raise NotImplementedError("This method is currently under development.")
    #
    # import pyspark.sql.functions as F
    #
    # # Validate inputs
    # if left_on or other_on:
    #     if not other_on:
    #         raise ValueError("If `left_on` is set, `other_on` should also be set")
    #     if not left_on:
    #         raise ValueError("If `other_on` is set, `left_on` should also be set")
    # if not (on or left_on):
    #     raise ValueError("Either `on` or (`left_on` and `other_on`) should be set")
    #
    # # Parse inputs
    # if on is None:
    #     on = []
    # elif isinstance(on, str):
    #     on = [on]
    # if left_on is None:
    #     left_on = []
    # elif isinstance(left_on, str):
    #     left_on = [left_on]
    # if other_on is None:
    #     other_on = []
    # elif isinstance(other_on, str):
    #     other_on = [other_on]
    #
    # wml = left_watermark
    # wmo = other_watermark
    # if wml is not None and not isinstance(wml, Watermark):
    #     wml = Watermark(**wml)
    # if wmo is not None and not isinstance(wmo, Watermark):
    #     wmo = Watermark(**wmo)
    #
    # # Set watermarks
    # if wml is None:
    #     try:
    #         wml = watermark(left)
    #     except Exception as e:
    #         logger.warn(f"Could not fetch wartermark from left dataframe: {e}")
    # else:
    #     left = left.withWatermark(wml.column, wml.threshold)
    # if wmo is None:
    #     try:
    #         wmo = watermark(other)
    #     except Exception as e:
    #         logger.warn(f"Could not fetch wartermark from other dataframe: {e}")
    # else:
    #     other = other.withWatermark(wmo.column, wmo.threshold)
    #
    # # Add watermark
    # other_cols = []
    # if wmo is not None:
    #     other_cols += [F.col(wmo.column).alias("_other_wc")]
    # other_cols += [F.col(c) for c in other.columns]
    # other = other.select(other_cols)
    #
    # # Drop duplicates to prevent adding rows to left
    # if on:
    #     other = other.dropDuplicates(on)
    #
    # _join = []
    # for c in on:
    #     _join += [f"left.{c} == other.{c}"]
    # for l, o in zip(left_on, other_on):
    #     _join += [f"left.{l} == other.{o}"]
    # if on_expression:
    #     _join += [on_expression]
    #
    # if wmo is not None:
    #     if time_constraint_interval_lower:
    #         _join += [
    #             f"left.{wml.column} >= other._other_wc - interval {time_constraint_interval_lower}"
    #         ]
    #     if time_constraint_interval_upper:
    #         _join += [
    #             f"left.{wml.column} <= other._other_wc + interval {time_constraint_interval_upper}"
    #         ]
    # _join = " AND ".join(_join)
    #
    # logger.info(f"   ON {_join}")
    # logger.debug(f"Left Schema: {left.schema}")
    # logger.debug(f"Other Schema: {other.schema}")
    #
    # df = self._df.alias("left").join(
    #     other=other.alias("other"),
    #     on=F.expr(_join),
    #     how=how,
    # )
    #
    # # Find duplicated columns (because of join)
    # d = defaultdict(lambda: 0)
    # for c in df.columns:
    #     d[c] += 1
    #
    # # Drop duplicated columns
    # for c, v in d.items():
    #     if v < 2:
    #         continue
    #
    #     if not (coalesce or c in _join):
    #         continue
    #     df = df.withColumn("__tmp", F.coalesce(f"left.{c}", f"other.{c}"))
    #     df = df.drop(c)
    #     df = df.withColumn(c, F.col("__tmp"))
    #     df = df.drop("__tmp")
    #
    # # Drop watermark column
    # if wmo is not None:
    #     df = df.drop(F.col("other._other_wc"))
    # logger.debug(f"Joined Schema: {df.schema}")
    #
    # return df
    #



import re
from typing import Any

import narwhals as nw
from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.models.readerwritermethod import ReaderWriterMethod

logger = get_logger(__name__)


class TableDataSource(BaseDataSource):
    catalog_name: str | None = Field(
        None,
        description="Source table catalog name",
    )
    schema_name: str | None = Field(
        None,
        description="Source table schema name",
    )
    table_name: str = Field(
        ...,
        description="""
        Source table name. Also supports fully qualified name (`{catalog}.{schema}.{table}`). In this case, 
        `catalog_name` and `schema_name` arguments are ignored.
        """,
    )
    reader_methods: list[ReaderWriterMethod] = Field(
        [], description="DataFrame backend reader methods."
    )

    @model_validator(mode="after")
    def table_full_name(self) -> Any:
        name = self.table_name
        if name is None:
            return

        # names = re.findall(r"\$\{vars\.[^}]+\}|[^.]+", name)
        names = re.split(r"\.(?![^{}]*})", name)

        with self.validate_assignment_disabled():
            self.table_name = names[-1]
            if len(names) > 1:
                self.schema_name = names[-2]
            if len(names) > 2:
                self.catalog_name = names[-3]

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        """Table full name {catalog_name}.{schema_name}.{table_name}"""
        if self.table_name is None:
            return None

        name = ""
        if self.catalog_name is not None:
            name = self.catalog_name

        if self.schema_name is not None:
            if name == "":
                name = self.schema_name
            else:
                name += f".{self.schema_name}"

        if name == "":
            name = self.table_name
        else:
            name += f".{self.table_name}"

        return name

    @property
    def _id(self) -> str:
        return self.full_name

    def _read_spark(self, spark=None) -> nw.LazyFrame:
        from laktory import get_spark_session

        spark = get_spark_session()

        if self.as_stream:
            logger.info(f"Reading {self._id} as stream")
            df = spark.readStream.table(self.full_name)
        else:
            logger.info(f"Reading {self._id} as static")
            df = spark.read.table(self.full_name)

        return nw.from_native(df)



from typing import Literal

from pydantic import Field

from laktory._logger import get_logger
from laktory.models.datasinks.tabledatasink import TableDataSink

logger = get_logger(__name__)


class HiveMetastoreDataSink(TableDataSink):
    """
    Data sink writing to a Hive Metastore data table.

    Examples
    ---------
    ```python
    import laktory as lk

    df = spark.createDataFrame([{"x": 1}, {"x": 2}, {"x": 3}])

    sink = lk.models.HiveMetastoreDataSink(
        schema_name="default",
        table_name="my_table",
        mode="APPEND",
    )
    # sink.write(df)
    ```
    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    # connection: Connection = None
    type: Literal["HIVE_METASTORE"] = Field(
        "HIVE_METASTORE", frozen=True, description="Sink Type"
    )

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Writers                                                                 #
    # ----------------------------------------------------------------------- #



import narwhals as nw


def row_index() -> nw.Expr:
    """
    Window function: returns a sequential number starting at 1 within a window partition.

    Returns
    -------
    :
        Output column

    Examples
    --------
    # ```py
    # import narwhals as nw
    # import polars as pl
    #
    # import laktory as lk  # noqa: F401
    #
    # df = nw.from_native(
    #     pl.DataFrame(
    #         {
    #             "x": ["a", "a", "b", "b", "b", "c"],
    #             "z": ["11", "10", "22", "21", "20", "30"],
    #         }
    #     )
    # )
    # df = df.with_columns(y1=nw.laktory.row_number())
    # df = df.with_columns(y2=nw.laktory.row_number().over("x"))
    # df = df.with_columns(y3=nw.laktory.row_number().over("x", sort_by="z"))
    # print(df)
    # '''
    # Rows: 6
    # Columns: 5
    # $ x  <str> 'a', 'a', 'b', 'b', 'b', 'c'
    # $ z  <str> '11', '10', '22', '21', '20', '30'
    # $ y1 <u32> 1, 2, 3, 4, 5, 6
    # $ y2 <u32> 1, 2, 1, 2, 3, 1
    # $ y3 <u32> 2, 1, 3, 2, 1, 1
    # '''
    # ```
    """

    # TODO: This one needs to be implemented as function in Narwhals and it's not
    # trivial. We might take a shot at implementing it or simply use df.with_row_index()
    # https://github.com/narwhals-dev/narwhals/issues/2307
    raise NotImplementedError()



from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline.pipeline import Pipeline
from laktory.models.resources.databricks.accesscontrolruleset import (
    AccessControlRuleSet,
)
from laktory.models.resources.databricks.alert import Alert
from laktory.models.resources.databricks.catalog import Catalog
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.clusterpolicy import ClusterPolicy
from laktory.models.resources.databricks.currentuser import CurrentUser
from laktory.models.resources.databricks.dashboard import Dashboard
from laktory.models.resources.databricks.dbfsfile import DbfsFile
from laktory.models.resources.databricks.directory import Directory
from laktory.models.resources.databricks.externallocation import ExternalLocation
from laktory.models.resources.databricks.grant import Grant
from laktory.models.resources.databricks.grants import Grants
from laktory.models.resources.databricks.group import Group
from laktory.models.resources.databricks.job import Job
from laktory.models.resources.databricks.metastore import Metastore
from laktory.models.resources.databricks.metastoreassignment import MetastoreAssignment
from laktory.models.resources.databricks.metastoredataaccess import MetastoreDataAccess
from laktory.models.resources.databricks.mlflowexperiment import MLflowExperiment
from laktory.models.resources.databricks.mlflowmodel import MLflowModel
from laktory.models.resources.databricks.mlflowwebhook import MLflowWebhook
from laktory.models.resources.databricks.mwsnetworkconnectivityconfig import (
    MwsNetworkConnectivityConfig,
)
from laktory.models.resources.databricks.notebook import Notebook
from laktory.models.resources.databricks.notificationdestination import (
    NotificationDestination,
)
from laktory.models.resources.databricks.obotoken import OboToken
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.databricks.pipeline import Pipeline as DatabricksPipeline
from laktory.models.resources.databricks.pythonpackage import PythonPackage
from laktory.models.resources.databricks.qualitymonitor import QualityMonitor
from laktory.models.resources.databricks.query import Query
from laktory.models.resources.databricks.recipient import Recipient
from laktory.models.resources.databricks.repo import Repo
from laktory.models.resources.databricks.schema import Schema
from laktory.models.resources.databricks.secret import Secret
from laktory.models.resources.databricks.secretscope import SecretScope
from laktory.models.resources.databricks.serviceprincipal import ServicePrincipal
from laktory.models.resources.databricks.share import Share
from laktory.models.resources.databricks.storagecredential import StorageCredential
from laktory.models.resources.databricks.table import Table
from laktory.models.resources.databricks.user import User
from laktory.models.resources.databricks.vectorsearchendpoint import (
    VectorSearchEndpoint,
)
from laktory.models.resources.databricks.vectorsearchindex import VectorSearchIndex
from laktory.models.resources.databricks.volume import Volume
from laktory.models.resources.databricks.warehouse import Warehouse
from laktory.models.resources.databricks.workspacebinding import WorkspaceBinding
from laktory.models.resources.databricks.workspacefile import WorkspaceFile
from laktory.models.resources.databricks.workspacetree import WorkspaceTree
from laktory.models.resources.providers.awsprovider import AWSProvider
from laktory.models.resources.providers.azureprovider import AzureProvider
from laktory.models.resources.providers.azurepulumiprovider import AzurePulumiProvider
from laktory.models.resources.providers.baseprovider import BaseProvider
from laktory.models.resources.providers.databricksprovider import DatabricksProvider

logger = get_logger(__name__)

DIRPATH = "./"


class Terraform(BaseModel):
    backend: Union[dict[str, Any], None] = None


class LaktorySettings(BaseModel):
    """
    Laktory Settings
    """

    dataframe_backend: str = Field(None, description="DataFrame backend")
    dataframe_api: Literal["NARWHALS", "NATIVE"] = Field(None, description="")
    workspace_laktory_root: str = Field(
        "/.laktory/",
        description="Root directory of a Databricks Workspace (excluding `'/Workspace') to which databricks objects like notebooks and workspace files are deployed.",
    )
    laktory_root: str = Field(
        "/laktory/",
        description="Laktory cache root directory. Used when a pipeline needs to write checkpoint files.",
    )

    @model_validator(mode="after")
    def apply_settings(self) -> Any:
        if self.dataframe_backend:
            settings.dataframe_backend = self.dataframe_backend

        if self.workspace_laktory_root:
            settings.workspace_laktory_root = self.workspace_laktory_root

        if self.laktory_root:
            settings.laktory_root = self.laktory_root

        if self.dataframe_api:
            settings.dataframe_api = self.dataframe_api

        return self


class Pulumi(BaseModel):
    """
    References
    ----------
    - Pulumi [configuration](https://www.pulumi.com/docs/concepts/config/)
    - Pulumi [outputs](https://www.pulumi.com/docs/concepts/inputs-outputs/#outputs)
    """

    config: dict[str, str] = Field(
        {},
        description="Pulumi configuration settings. Generally used to configure providers. See references for more details.",
    )
    outputs: dict[str, str] = Field(
        {},
        description="Requested resources-related outputs. See references for details.",
    )


class StackResources(BaseModel):
    """
    Resources definition for a given stack or stack environment.
    """

    databricks_accesscontrolrulesets: dict[str, AccessControlRuleSet] = {}
    databricks_alerts: dict[str, Alert] = {}
    databricks_catalogs: dict[str, Catalog] = {}
    databricks_clusterpolicies: dict[str, ClusterPolicy] = {}
    databricks_clusters: dict[str, Cluster] = {}
    databricks_currentusers: dict[str, CurrentUser] = {}
    databricks_dashboards: dict[str, Dashboard] = {}
    databricks_dbfsfiles: dict[str, DbfsFile] = {}
    databricks_directories: dict[str, Directory] = {}
    databricks_pipelines: dict[str, DatabricksPipeline] = {}
    databricks_externallocations: dict[str, ExternalLocation] = {}
    databricks_grant: dict[str, Grant] = {}
    databricks_grants: dict[str, Grants] = {}
    databricks_groups: dict[str, Group] = {}
    databricks_jobs: dict[str, Job] = {}
    databricks_metastoreassignments: dict[str, MetastoreAssignment] = {}
    databricks_metastoredataaccesses: dict[str, MetastoreDataAccess] = {}
    databricks_metastores: dict[str, Metastore] = {}
    databricks_mlflowexperiments: dict[str, MLflowExperiment] = {}
    databricks_mlflowmodels: dict[str, MLflowModel] = {}
    databricks_mlflowwebhooks: dict[str, MLflowWebhook] = {}
    databricks_networkconnectivityconfig: dict[str, MwsNetworkConnectivityConfig] = {}
    databricks_notebooks: dict[str, Notebook] = {}
    databricks_notificationdestinations: dict[str, NotificationDestination] = {}
    databricks_obotokens: dict[str, OboToken] = {}
    databricks_permissions: dict[str, Permissions] = {}
    databricks_qualitymonitors: dict[str, QualityMonitor] = {}
    databricks_pythonpackages: dict[str, PythonPackage] = {}
    databricks_queries: dict[str, Query] = {}
    databricks_recipients: dict[str, Recipient] = {}
    databricks_repos: dict[str, Repo] = {}
    databricks_schemas: dict[str, Schema] = {}
    databricks_secrets: dict[str, Secret] = {}
    databricks_secretscopes: dict[str, SecretScope] = {}
    databricks_serviceprincipals: dict[str, ServicePrincipal] = {}
    databricks_shares: dict[str, Share] = {}
    databricks_storagecredentials: dict[str, StorageCredential] = {}
    databricks_tables: dict[str, Table] = {}
    databricks_users: dict[str, User] = {}
    databricks_vectorsearchendpoints: dict[str, VectorSearchEndpoint] = {}
    databricks_vectorsearchindexes: dict[str, VectorSearchIndex] = {}
    databricks_volumes: dict[str, Volume] = {}
    databricks_warehouses: dict[str, Warehouse] = {}
    databricks_workspacebindings: dict[str, WorkspaceBinding] = {}
    databricks_workspacefiles: dict[str, WorkspaceFile | PythonPackage] = {}
    databricks_workspacetrees: dict[str, WorkspaceTree | PythonPackage] = {}
    pipelines: dict[str, Pipeline] = {}
    providers: dict[
        str, Union[AWSProvider, AzureProvider, AzurePulumiProvider, DatabricksProvider]
    ] = {}

    @model_validator(mode="after")
    def update_resource_names(self) -> Any:
        for k, r in self._get_all().items():
            if r.resource_name_ and k != r.resource_name_:
                raise ValueError(
                    f"Provided resource name {r.resource_name_} does not match provided key {k}"
                )
            r.resource_name_ = k
        return self

    def _get_all(self, providers_excluded=False, providers_only=False):
        resources = {}
        for resource_type in self.model_fields.keys():
            if resource_type in ["variables"]:
                continue

            for resource_name, _r in getattr(self, resource_type).items():
                if resource_name in resources.keys():
                    raise ValueError(
                        f"Stack resource names are not unique. '{resource_name}' is already used."
                    )

                if providers_excluded and isinstance(_r, BaseProvider):
                    continue

                if providers_only and not isinstance(_r, BaseProvider):
                    continue

                resources[resource_name] = _r

        return resources


class EnvironmentSettings(BaseModel):
    """
    Settings overwrite for a specific environments
    """

    resources: Any = Field(
        None,
        description="""
    Dictionary of resources to be deployed. Each key should be a resource type and each value should be a dictionary of
    resources who's keys are the resource names and the values the resources definitions.
    """,
    )
    variables: dict[str, Any] = Field(
        None,
        description="Dictionary of variables made available in the resources definition.",
    )
    terraform: Terraform = Field(Terraform(), description="Terraform-specific settings")


class Stack(BaseModel):
    """
    The Stack defines a collection of deployable resources, the deployment
    configuration, some variables and the environment-specific settings.

    Examples
    --------
    ```py
    from laktory import models

    stack = models.Stack(
        name="workspace",
        backend="pulumi",
        pulumi={
            "config": {
                "databricks:host": "${vars.DATABRICKS_HOST}",
                "databricks:token": "${vars.DATABRICKS_TOKEN}",
            },
        },
        resources={
            "databricks_pipelines": {
                "pl-stock-prices": {
                    "name": "pl-stock-prices",
                    "development": "${vars.is_dev}",
                    "libraries": [
                        {"notebook": {"path": "/pipelines/dlt_brz_template.py"}},
                    ],
                }
            },
            "databricks_jobs": {
                "job-stock-prices": {
                    "name": "job-stock-prices",
                    "job_clusters": [
                        {
                            "job_cluster_key": "main",
                            "new_cluster": {
                                "spark_version": "16.3.x-scala2.12",
                                "node_type_id": "Standard_DS3_v2",
                            },
                        }
                    ],
                    "tasks": [
                        {
                            "task_key": "ingest",
                            "job_cluster_key": "main",
                            "notebook_task": {
                                "notebook_path": "/.laktory/jobs/ingest_stock_prices.py",
                            },
                        },
                        {
                            "task_key": "pipeline",
                            "depends_ons": [{"task_key": "ingest"}],
                            "pipeline_task": {
                                "pipeline_id": "${resources.dlt-pl-stock-prices.id}",
                            },
                        },
                    ],
                }
            },
        },
        variables={
            "org": "okube",
        },
        environments={
            "dev": {
                "variables": {
                    "is_dev": True,
                }
            },
            "prod": {
                "variables": {
                    "is_dev": False,
                }
            },
        },
    )
    ```
    References
    ----------
    * [Stack](https://www.laktory.ai/concepts/stack/)
    """

    backend: Literal["pulumi", "terraform"] = Field(
        None, description="IaC backend used for deployment."
    )
    description: str = Field(None, description="Description of the stack")
    environments: dict[str, EnvironmentSettings] = Field(
        {},
        description="Environment-specific overwrite of config, resources or variables arguments.",
    )
    name: str = Field(
        ...,
        description="Name of the stack. If Pulumi is used as a backend, it should match the name of the Pulumi project.",
    )
    organization: Union[str, None] = Field(None, description="Organization")
    pulumi: Pulumi = Field(Pulumi(), description="Pulumi-specific settings")
    resources: Union[StackResources, None] = Field(
        StackResources(),
        description="""
    Dictionary of resources to be deployed. Each key should be a resource type and each value should be a dictionary of
    resources who's keys are the resource names and the values the resources definitions.
    """,
    )
    settings: LaktorySettings = Field(None, description="Laktory settings")
    terraform: Terraform = Field(Terraform(), description="Terraform-specific settings")
    variables: dict[str, Any] = Field(
        {},
        description="Dictionary of variables made available in the resources definition.",
    )

    @model_validator(mode="before")
    @classmethod
    def apply_settings(cls, data: Any) -> Any:
        """Required to apply settings before instantiating resources and setting default values"""
        settings = data.get("settings", None)
        if settings:
            if not isinstance(settings, dict):
                settings = settings.model_dump(exclude_unset=True)
            LaktorySettings(**settings)

        return data

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def build(self, env_name: str | None, inject_vars: bool = True):
        """
        Build stack artifacts before preview or deploy.

        Parameters
        ----------
        env_name:
            Name of the environment
        inject_vars:
            Inject stack variables
        """

        logger.info("Building artifacts...")

        env = self.get_env(env_name=env_name)
        if inject_vars:
            env = env.inject_vars()

        for k, r in env.resources._get_all(providers_excluded=True).items():
            if isinstance(r, PythonPackage):
                r.build()

        logger.info("Writing pipeline config files...")
        for k, r in env.resources._get_all(providers_excluded=True).items():
            if isinstance(r, Pipeline):
                if not r.orchestrator:
                    continue
                config_file = getattr(r.orchestrator, "config_file")
                if config_file:
                    config_file.build()

        logger.info("Build completed.")

    def get_env(self, env_name: str | None):
        """
        Complete definition the stack for a given environment. It takes into
        account both the default stack values and environment-specific
        overwrites.

        Parameters
        ----------
        env_name:
            Name of the environment

        Returns
        -------
        :
            Environment definitions.
        """

        if env_name is None:
            return self

        if env_name not in self.environments.keys():
            raise ValueError(f"Environment '{env_name}' is not declared in the stack.")

        env = self.model_copy(update={"environments": {}})
        env.update(self.environments[env_name].model_dump(exclude_unset=True))
        return env

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    def to_pulumi(self, env_name: Union[str, None] = None):
        """
        Create a pulumi stack for a given environment `env`.

        Parameters
        ----------
        env_name:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : PulumiStack
            Pulumi-specific stack definition
        """
        from laktory.models.stacks.pulumistack import PulumiStack

        env = self.get_env(env_name=env_name).inject_vars()
        env.build(env_name=None, inject_vars=False)

        # Resources
        resources = {}
        for r in env.resources._get_all().values():
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        return PulumiStack(
            name=env.name,
            organization=env.organization,
            config=env.pulumi.config,
            description=env.description,
            resources=resources,
            outputs=env.pulumi.outputs,
        )

    # ----------------------------------------------------------------------- #
    # Terraform Methods                                                       #
    # ----------------------------------------------------------------------- #

    def to_terraform(self, env_name: Union[str, None] = None):
        """
        Create a terraform stack for a given environment `env`.

        Parameters
        ----------
        env_name:
            Target environment. If `None`, used default stack values only.

        Returns
        -------
        : TerraformStack
            Terraform-specific stack definition
        """
        from laktory.models.stacks.terraformstack import TerraformStack

        env = self.get_env(env_name=env_name).inject_vars()
        env.build(env_name=None, inject_vars=False)

        # Providers
        providers = {}
        for r in env.resources._get_all(providers_only=True).values():
            for _r in r.core_resources:
                rname = _r.resource_name
                providers[rname] = _r

        # Resources
        resources = {}
        for r in env.resources._get_all(providers_excluded=True).values():
            for _r in r.core_resources:
                resources[_r.resource_name] = _r

        # Update terraform
        return TerraformStack(
            terraform={"backend": env.terraform.backend},
            providers=providers,
            resources=resources,
        )



import json
from typing import Any
from typing import Union

from pydantic import Field
from pydantic import field_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ClusterPolicyLibraryCran(BaseModel):
    package: str = Field(..., description="")
    repo: str = Field(None, description="")


class ClusterPolicyLibraryMaven(BaseModel):
    coordinates: str = Field(..., description="")
    exclusions: list[str] = Field(None, description="")
    repo: str = Field(None, description="")


class ClusterPolicyLibraryPypi(BaseModel):
    package: str = Field(None, description="")
    repo: str = Field(None, description="")


class ClusterPolicyLibrary(BaseModel):
    cran: ClusterPolicyLibraryCran = Field(
        None, description="Cran library specifications"
    )
    egg: str = Field(None, description="Egg filepath")
    jar: str = Field(None, description="Jar filepath")
    maven: ClusterPolicyLibraryMaven = Field(None, description="")
    pypi: ClusterPolicyLibraryPypi = Field(
        None, description="Pypi library specifications"
    )
    requirement: str = Field(None, description="")
    whl: str = Field(None, description="Wheel filepath")


class ClusterPolicy(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks cluster policy

    Examples
    --------
    ```py
    from laktory import models

    cluster = models.resources.databricks.ClusterPolicy(
        name="okube",
        definition={
            "dbus_per_hour": {
                "type": "range",
                "maxValue": 10,
            },
            "autotermination_minutes": {"type": "fixed", "value": 30, "hidden": True},
            "custom_tags.team": {
                "type": "fixed",
                "value": "okube",
            },
        },
        libraries=[
            {
                "pypi": {
                    "package": "laktory==0.5.0",
                }
            }
        ],
        access_controls=[
            {"permission_level": "CAN_USE", "group_name": "account users"}
        ],
    )
    ```

    References
    ----------

    * [Databricks Cluster Policy](https://docs.databricks.com/api/workspace/clusterpolicies/create)
    * [Pulumi Databricks Cluster Policy](https://www.pulumi.com/registry/packages/databricks/api-docs/clusterpolicy)

    """

    access_controls: list[AccessControl] = Field(
        [], description="List of access controls"
    )
    definition: Union[str, dict[str, Any]] = Field(
        None,
        description="""
    Policy definition: JSON document expressed in [Databricks Policy Definition Language](https://docs.databricks.com/en/admin/clusters/policies.html#cluster-policy-definition).
    Cannot be used with `policy_family_id`.
    """,
    )
    description: str = Field(
        None, description="Additional human-readable description of the cluster policy."
    )
    libraries: list[ClusterPolicyLibrary] = Field(None, description="")
    max_clusters_per_user: int = Field(
        None,
        description="""
    Maximum number of clusters allowed per user. When omitted, there is no limit. If specified, value must be greater 
    than zero.
    """,
    )
    name: str = Field(
        ...,
        description="Cluster policy name. This must be unique. Length must be between 1 and 100 characters.",
    )
    policy_family_definition_overrides: str = Field(
        None,
        description="""
    Policy definition JSON document expressed in Databricks Policy Definition Language. The JSON document must be 
    passed as a string and cannot be embedded in the requests. You can use this to customize the policy definition 
    inherited from the policy family. Policy rules specified here are merged into the inherited policy definition.
    """,
    )
    policy_family_id: str = Field(
        None,
        description="""
    ID of the policy family. The cluster policy's policy definition inherits the policy family's policy definition.
    Cannot be used with `definition`. Use `policy_family_definition_overrides` instead to customize the policy 
    definition.
    """,
    )

    @field_validator("definition")
    def validate_type(cls, v: Union[str, dict[str, str]]) -> str:
        if isinstance(v, dict):
            v = json.dumps(v)
        return v

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    cluster_policy_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:ClusterPolicy"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "libraries": "libraries",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_cluster_policy"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import narwhals as nw


def string_split(
    self,
    pattern: str,
    key: int,
) -> nw.Expr:
    """
    Get substring using separator `pattern`.

    Parameters
    ----------
    pattern:
        String or regular expression to split on. If not specified, split on whitespace.
    key:
        Split index to return

    Returns
    -------
    :
        Result

    Examples
    --------
    # ```py
    # import polars as pl
    #
    # import laktory as lk  # noqa: F401
    #
    # df = nw.from_native(pl.DataFrame({"x": ["price_close"]}))
    #
    # df = df.with_columns(y=nw.col("x").laktory.string_split(pattern="_", key=1))
    # print(df.to_native().glimpse(return_as_string=True))
    # '''
    # Rows: 1
    # Columns: 2
    # $ x <str> 'price_close'
    # $ y <str> 'close'
    # '''
    ```
    """
    raise NotImplementedError()
    return self._expr.str.split(by=pattern).list.get(key, null_on_oob=True)



import time
from typing import TYPE_CHECKING
from typing import Literal

from laktory._logger import get_logger
from laktory.dispatcher.dispatcherrunner import DispatcherRunner

if TYPE_CHECKING:
    from databricks.sdk.service.jobs import Run
    from databricks.sdk.service.jobs import Wait

logger = get_logger(__name__)


class DatabricksJobRunner(DispatcherRunner):
    """
    Job runner.
    """

    _run_start: "Wait" = None
    _run: "Run" = None

    def get_id(self) -> str:
        """Get deployed job id"""
        logger.info(f"Getting id for job {self.name}")
        for _job in self.wc.jobs.list(name=self.name):
            _job_data = _job.as_dict()
            if self.name == _job_data["settings"]["name"]:
                self.id = _job.job_id
                return self.id

    def run(
        self,
        wait: bool = True,
        timeout: int = 20 * 60,
        raise_exception: bool = False,
        current_run_action: Literal["WAIT", "CANCEL", "FAIL"] = "WAIT",
    ):
        """
        Run remote job and monitor failures.

        Parameters
        ----------
        wait:
            If `True`, runs synchronously and wait for completion.
        timeout:
            Maximum time allowed for the job to run before an exception is
            raised.
        raise_exception:
            If `True`, exceptions are raised if the job fails.
        current_run_action:
            Action to take with respect to current run (if any). Possible
            options are:
                - WAIT: wait for the current run to complete
                - CANCEL: cancel the current run
                - FAIL: raise an exception

        Returns
        -------
        output:
            None
        """
        from databricks.sdk.errors import OperationFailed
        from databricks.sdk.service.jobs import RunLifeCycleState

        active_runs = list(self.wc.jobs.list_runs(job_id=self.id, active_only=True))

        if len(active_runs) > 0:
            if current_run_action.upper() == "FAIL":
                logger.info(f"Job {self.name} already running...")
                raise Exception(f"Job {self.name} already running...")

            elif current_run_action.upper() == "WAIT":
                logger.info(
                    f"Job {self.name} waiting for current run(s) to be completed..."
                )
                for run in active_runs:
                    try:
                        self.wc.jobs.wait_get_run_job_terminated_or_skipped(
                            run_id=run.run_id
                        )
                    except OperationFailed:
                        pass

            elif current_run_action.upper() == "CANCEL":
                logger.info(f"Job {self.name} cancelling current run(s)...")
                for run in active_runs:
                    self.wc.jobs.cancel_run_and_wait(run_id=run.run_id)

        # Start update
        t0 = time.time()
        logger.info(f"Job {self.name} run started...")
        self._run_start = self.wc.jobs.run_now(
            job_id=self.id,
        )

        pstates = {}
        self.get_run()
        logger.info(f"Job {self.name} run URL: {self._run.run_page_url}")
        if wait:
            while (
                time.time() - t0 < timeout
                or self.run_state == RunLifeCycleState.TERMINATED
            ):
                self.get_run()

                if self.run_state != pstates.get(self.run_id, None):
                    logger.info(f"Job {self.name} state: {self.run_state.value}")
                    pstates[self.run_id] = self.run_state

                for task in self._run.tasks:
                    state = task.state
                    if state.life_cycle_state != pstates.get(task.run_id, None):
                        logger.info(
                            f"   Task {self.name}.{task.task_key} state: {state.life_cycle_state.value}"
                        )
                    pstates[task.run_id] = state.life_cycle_state

                if self.run_state in [
                    RunLifeCycleState.TERMINATED,
                    RunLifeCycleState.SKIPPED,
                    RunLifeCycleState.INTERNAL_ERROR,
                ]:
                    break

                time.sleep(1)

            logger.info(
                f"Job {self.name} run terminated after {time.time() - t0: 5.2f} sec with {self.run_state} ({self._run.state.state_message})"
            )
            for task in self._run.tasks:
                logger.info(
                    f"Task {self.name}.{task.task_key} terminated with {task.state.result_state} ({task.state.state_message})"
                )
            if raise_exception and self.run_state != RunLifeCycleState.TERMINATED:
                raise Exception(
                    f"Job {self.name} update not completed ({self.run_state})"
                )

    def get_run(self):
        self._run = self.wc.jobs.get_run(
            run_id=self.run_id,
        )
        return self._run

    @property
    def run_state(self):
        return self._run.state.life_cycle_state

    @property
    def run_id(self) -> str:
        return self._run_start.run_id



# Databricks notebook source
# COMMAND ----------
import json

reqs = spark.conf.get("requirements")
reqs = " ".join(json.loads(reqs))
# MAGIC %pip install $reqs
# MAGIC %restart_python

# COMMAND ----------
import dlt  # noqa: E402
import pyspark.sql.functions as F  # noqa: F401, E402

import laktory as lk  # noqa: E402

# --------------------------------------------------------------------------- #
# Read Pipeline                                                               #
# --------------------------------------------------------------------------- #

config_filepath = spark.conf.get("config_filepath")
print(f"Reading pipeline at {config_filepath}")
with open(config_filepath, "r") as fp:
    pl = lk.models.Pipeline.model_validate_json(fp.read())

# --------------------------------------------------------------------------- #
# Tables and Views Definition                                                 #
# --------------------------------------------------------------------------- #


def define_table(node, sink):
    table_or_view = dlt.table
    if isinstance(sink, lk.models.PipelineViewDataSink):
        table_or_view = dlt.view

    if not sink.is_cdc:

        @table_or_view(**sink.dlt_table_or_view_kwargs)
        @dlt.expect_all(sink.dlt_warning_expectations)
        @dlt.expect_all_or_drop(sink.dlt_drop_expectations)
        @dlt.expect_all_or_fail(sink.dlt_fail_expectations)
        def get_df():
            # Execute node
            node.execute()
            if sink.is_quarantine:
                df = node.quarantine_df
            else:
                df = node.output_df

            # Return
            return df.to_native()

    else:

        @dlt.view(name=sink.dlt_pre_merge_view_name)
        def get_df():
            node.execute()
            return node.output_df.to_native()

        dlt.create_streaming_table(**sink.dlt_table_or_view_kwargs)
        dlt.apply_changes(**sink.dlt_apply_changes_kwargs)


# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

# Build nodes
for node in pl.nodes:
    if node.dlt_template != "DEFAULT":
        continue

    for sink in node.sinks:
        define_table(node, sink)



import hashlib
import os
import shutil
import uuid
from pathlib import Path
from typing import Any
from typing import Literal

import narwhals as nw
from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field
from pydantic import field_serializer
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.basemodel import BaseModel
from laktory.models.datasinks.mergecdcoptions import DataSinkMergeCDCOptions
from laktory.models.pipelinechild import PipelineChild
from laktory.models.readerwritermethod import ReaderWriterMethod
from laktory.typing import AnyFrame

logger = get_logger(__name__)


SUPPORTED_BACKENDS = [DataFrameBackends.POLARS, DataFrameBackends.PYSPARK]
LAKTORY_MODES = ["MERGE"]
SPARK_MODES = [
    "OVERWRITE",
    "APPEND",
    "IGNORE",
    "ERROR",
    "ERRORIFEXISTS",
] + LAKTORY_MODES
SPARK_STREAMING_MODES = ["APPEND", "COMPLETE", "UPDATE"] + LAKTORY_MODES
POLARS_DELTA_MODES = ["ERROR", "APPEND", "OVERWRITE"] + LAKTORY_MODES
SUPPORTED_MODES = tuple(set(SPARK_MODES + SPARK_STREAMING_MODES + POLARS_DELTA_MODES))


class BaseDataSink(BaseModel, PipelineChild):
    """Base class for data sinks"""

    databricks_quality_monitor: Literal[None] = Field(
        None, description="Databricks Quality Monitor"
    )
    checkpoint_path_: str | Path = Field(
        None,
        description="Path to which the checkpoint file for which a streaming dataframe should be written.",
        validation_alias=AliasChoices("checkpoint_path", "checkpoint_path_"),
        exclude=True,
    )
    is_quarantine: bool = Field(
        False,
        description="Sink used to store quarantined results from a pipeline node expectations.",
    )
    type: Literal["FILE", "HIVE_METASTORE", "UNITY_CATALOG"] = Field(
        ..., description="Name of the data sink type"
    )
    metadata: Literal[None] = Field(None, description="Table and columns metadata.")
    merge_cdc_options: DataSinkMergeCDCOptions = Field(
        None,
        description="Merge options to handle input DataFrames that are Change Data Capture (CDC). Only used when `MERGE` mode is selected.",
    )  # TODO: Review parameter name
    mode: Literal.__getitem__(SUPPORTED_MODES) | None = Field(
        None,
        description="""
        Write mode.
        
        Spark
        -----
        - OVERWRITE: Overwrite existing data.
        - APPEND: Append contents of this DataFrame to existing data.
        - ERROR: Throw an exception if data already exists.
        - IGNORE: Silently ignore this operation if data already exists.

        Spark Streaming
        ---------------
        - APPEND: Only the new rows in the streaming DataFrame/Dataset will be written to the sink.
        - COMPLETE: All the rows in the streaming DataFrame/Dataset will be written to the sink every time there are some updates.
        - UPDATE: Only the rows that were updated in the streaming DataFrame/Dataset will be written to the sink every time there are some updates.

        Polars Delta
        ------------
        - OVERWRITE: Overwrite existing data.
        - APPEND: Append contents of this DataFrame to existing data.
        - ERROR: Throw an exception if data already exists.
        - IGNORE: Silently ignore this operation if data already exists.

        Laktory
        -------
        - MERGE: Append, update and optionally delete records. Only supported for DELTA format. Requires cdc specification.
        """,
    )
    writer_kwargs: dict[str, Any] = Field(
        {},
        description="Keyword arguments passed directly to dataframe backend writer. Passed to `.options()` method when using PySpark.",
    )
    writer_methods: list[ReaderWriterMethod] = Field(
        [], description="DataFrame backend writer methods."
    )

    @model_validator(mode="after")
    def merge_has_options(self) -> Any:
        if self.mode == "MERGE":
            if self.merge_cdc_options is None:
                raise ValueError(
                    "If 'MERGE' `mode` is selected, `merge_cdc_options` must be specified."
                )
            else:
                self.merge_cdc_options._parent = self

            from laktory.models.pipeline.orchestrators.databrickspipelineorchestrator import (
                DatabricksPipelineOrchestrator,
            )

            if (
                self.parent_pipeline
                and self.parent_pipeline.orchestrator
                and isinstance(
                    self.parent_pipeline.orchestrator, DatabricksPipelineOrchestrator
                )
            ):
                if self.merge_cdc_options.order_by is None:
                    raise ValueError(
                        "`merge_cdc_options` requires a value for `order_by` when Databricks Pipeline orchestrator is selected."
                    )

        return self

    # -------------------------------------------------------------------------------- #
    # Properties                                                                       #
    # -------------------------------------------------------------------------------- #

    @property
    def _id(self):
        raise NotImplementedError()

    @property
    def _uuid(self) -> str:
        _id = self._id
        hash_object = hashlib.sha1(_id.encode())
        hash_digest = hash_object.hexdigest()
        return str(uuid.UUID(hash_digest[:32]))

    @computed_field(description="checkpoint_path")
    @property
    def checkpoint_path(self) -> Path | None:
        if self.checkpoint_path_:
            return Path(self.checkpoint_path_)

        node = self.parent_pipeline_node

        if node and node.root_path:
            for i, s in enumerate(node.all_sinks):
                if s == self:
                    return node.root_path / "checkpoints" / f"sink-{self._uuid}"
        return None

    @field_serializer("checkpoint_path", when_used="json")
    def serialize_path(self, value: Path) -> str:
        return value.as_posix()

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to write sink"""
        return []

    @property
    def data_sources(self):
        """Get all sources feeding the sink"""
        return []

    # -------------------------------------------------------------------------------- #
    # CDC                                                                              #
    # -------------------------------------------------------------------------------- #

    @property
    def is_cdc(self) -> bool:
        return self.merge_cdc_options is not None

    @property
    def dlt_pre_merge_view_name(self):
        """
        DLT view applying node transformer prior to applying CDC changes.
        """
        return "_" + self.dlt_table_or_view_name.split(".")[-1]

    @property
    def dlt_apply_changes_kwargs(self) -> dict[str, str]:
        """
        Keyword arguments for dlt.apply_changes function
        """

        # if not isinstance(self, TableDataSink):
        #     raise ValueError("DLT only supports `TableDataSink` class")

        cdc = self.merge_cdc_options
        return {
            "apply_as_deletes": cdc.delete_where,
            # "apply_as_truncates": ,  # NOT SUPPORTED
            "column_list": cdc.include_columns,
            "except_column_list": cdc.exclude_columns,
            "ignore_null_updates": cdc.ignore_null_updates,
            "keys": cdc.primary_keys,
            "sequence_by": cdc.order_by,
            "source": self.dlt_pre_merge_view_name,
            "stored_as_scd_type": cdc.scd_type,
            "target": self.table_name,
            # "track_history_column_list": cdc.track_history_columns,  # NOT SUPPORTED
            # "track_history_except_column_list": cdc.track_history_except_columns,  # NOT SUPPORTED
        }

    # -------------------------------------------------------------------------------- #
    # Writers                                                                          #
    # -------------------------------------------------------------------------------- #

    def _validate_format(self):
        return None

    def _validate_mode(self, mode, df):
        if self.dataframe_backend == DataFrameBackends.POLARS:
            self._validate_mode_polars(mode, df)
        elif self.dataframe_backend == DataFrameBackends.PYSPARK:
            self._validate_mode_spark(mode, df)

    def write(
        self, df: AnyFrame = None, mode: str = None, full_refresh: bool = False
    ) -> None:
        """
        Write dataframe into sink.

        Parameters
        ----------
        df:
            Input dataframe.
        full_refresh
            If `True`, source is deleted/dropped (including checkpoint if applicable)
            before write.
        mode:
            Write mode overwrite of the sink default mode.
        """

        if getattr(self, "view_definition", None):
            if self.dataframe_backend == DataFrameBackends.PYSPARK:
                self._write_spark_view()
            elif self.dataframe_backend == DataFrameBackends.POLARS:
                self._write_polars_view()
            else:
                raise ValueError(
                    f"DataFrame backend '{self.dataframe_backend}' is not supported"
                )
            return

        if mode is None:
            mode = self.mode

        if not isinstance(df, (nw.DataFrame, nw.LazyFrame)):
            df = nw.from_native(df)

        dataframe_backend = DataFrameBackends.from_nw_implementation(df.implementation)
        if dataframe_backend not in SUPPORTED_BACKENDS:
            raise ValueError(
                f"DataFrame provided is of {dataframe_backend} backend, which is not supported."
            )

        if self.dataframe_backend_ and self.dataframe_backend_ != dataframe_backend:
            raise ValueError(
                f"DataFrame provided is {dataframe_backend} and source has been configure with {self.dataframe_backend_} backend."
            )
        self.dataframe_backend_ = dataframe_backend

        self._validate_mode(mode, df)
        self._validate_format()

        if mode and mode.lower() == "merge":
            self.merge_cdc_options.execute(source=df)
            logger.info("Write completed.")
            return

        if self.dataframe_backend == DataFrameBackends.PYSPARK:
            self._write_spark(df=df, mode=mode, full_refresh=full_refresh)
        elif self.dataframe_backend == DataFrameBackends.POLARS:
            self._write_polars(df=df, mode=mode, full_refresh=full_refresh)
        else:
            raise ValueError(
                f"DataFrame backend '{self.dataframe_backend}' is not supported"
            )

        logger.info("Write completed.")

    def _write_spark_view(self) -> None:
        raise NotImplementedError(
            f"View creation with spark is not implemented for type '{type(self)}'"
        )

    def _write_polars_view(self) -> None:
        raise NotImplementedError(
            f"View creation with polars is not implemented for type '{type(self)}'"
        )

    # -------------------------------------------------------------------------------- #
    # Writers - Spark                                                                  #
    # -------------------------------------------------------------------------------- #

    def _validate_mode_spark(self, mode, df):
        if df.to_native().isStreaming:
            if mode not in SPARK_STREAMING_MODES:
                raise ValueError(
                    f"Mode '{mode}' is not supported for Spark Streaming DataFrame. Choose from {SPARK_STREAMING_MODES}"
                )
        else:
            if mode not in SPARK_MODES:
                raise ValueError(
                    f"Mode '{mode}' is not supported for Spark DataFrame. Choose from {SPARK_MODES}"
                )

    def _get_spark_kwargs(self, mode, is_streaming):
        fmt = self.format.lower()

        # Build kwargs
        kwargs = {}

        kwargs["mergeSchema"] = True
        kwargs["overwriteSchema"] = False
        if mode in ["OVERWRITE", "COMPLETE"]:
            kwargs["mergeSchema"] = False
            kwargs["overwriteSchema"] = True
        if is_streaming:
            kwargs["checkpointLocation"] = self.checkpoint_path.as_posix()

        if fmt in ["jsonl", "ndjson"]:
            fmt = "json"
            kwargs["multiline"] = False
        elif fmt in ["json"]:
            kwargs["multiline"] = True

        # User Options
        for k, v in self.writer_kwargs.items():
            kwargs[k] = v

        return kwargs, fmt

    def _get_spark_writer_methods(self, mode, is_streaming):
        methods = []

        options, fmt = self._get_spark_kwargs(mode=mode, is_streaming=is_streaming)

        methods += [ReaderWriterMethod(name="format", args=[fmt])]

        if is_streaming:
            methods += [ReaderWriterMethod(name="outputMode", args=[mode])]
            methods += [
                ReaderWriterMethod(name="trigger", kwargs={"availableNow": True})
            ]
        else:
            methods += [ReaderWriterMethod(name="mode", args=[mode])]

        if options:
            methods += [ReaderWriterMethod(name="options", kwargs=options)]

        for m in self.writer_methods:
            methods += [m]

        return methods

    def _write_spark(self, df, mode, full_refresh) -> None:
        raise NotImplementedError(
            f"`{self.dataframe_backend}` not supported for `{type(self)}`"
        )

    # -------------------------------------------------------------------------------- #
    # Writers - Polars                                                                 #
    # -------------------------------------------------------------------------------- #

    def _validate_mode_polars(self, mode, df):
        if mode is not None:
            raise ValueError(
                f"Mode '{mode}' is not supported for Polars DataFrame. Set to `None`"
            )

    def _get_polars_kwargs(self, mode):
        fmt = self.format.lower()

        kwargs = {}

        if self.format != "DELTA":
            if mode:
                raise ValueError(
                    "'mode' configuration with Polars only supported by 'DELTA' format"
                )
        else:
            # if full_refresh or not self.exists():
            #     mode = "OVERWRITE"

            if not mode:
                raise ValueError(
                    "'mode' configuration required with Polars 'DELTA' format"
                )

        if mode:
            kwargs["mode"] = mode

        for k, v in self.writer_kwargs.items():
            kwargs[k] = v

        return kwargs, fmt

    def _write_polars(self, df, mode, full_refresh) -> None:
        raise NotImplementedError(
            f"`{self.dataframe_backend}` not supported for `{type(self)}`"
        )

    # ----------------------------------------------------------------------- #
    # Purge                                                                   #
    # ----------------------------------------------------------------------- #

    def exists(self):
        if self.dataframe_backend == DataFrameBackends.PYSPARK:
            from laktory import get_spark_session

            try:
                spark = get_spark_session()
                df = self.read(spark=spark, as_stream=False)
                df.limit(1).collect()
                return True
            except Exception:
                return False

        else:
            raise NotImplementedError()

    def _purge_checkpoint(self):
        if self.checkpoint_path:
            # Try with simple paths (supported by local file system or Unity Catalog)
            if os.path.exists(self.checkpoint_path):
                logger.info(
                    f"Deleting checkpoint at {self.checkpoint_path}",
                )
                shutil.rmtree(self.checkpoint_path)

            # Try with DBFS
            # If spark is not used, dbfs is most likely not used
            if self.dataframe_backend != DataFrameBackends.PYSPARK:
                return

            # Check if a workspace client can be instantiated
            try:
                from databricks.sdk import WorkspaceClient
                from databricks.sdk.errors import ResourceDoesNotExist

                w = WorkspaceClient()

            except (
                ModuleNotFoundError,  # SDK not installed
                ImportError,  # SDK with different version / API
                ValueError,  # client not configure (would never happen in a notebook)
            ):
                return

            # Format path for DBFS
            _path = self.checkpoint_path.as_posix()
            if not _path.startswith("/") and not _path.startswith("dbfs:"):
                _path = "/" + _path
            if _path.startswith("/dbfs/"):
                _path = _path.replace("/dbfs/", "dbfs:/")
            if not _path.startswith("dbfs:"):
                _path = "dbfs:" + _path

            # Check Status
            try:
                w.dbfs.get_status(_path)
            except ResourceDoesNotExist:
                logger.info(
                    f"Checkpoint at {_path} does not exist. Skipping.",
                )
                return

            logger.info(
                f"Deleting checkpoint at {_path}.",
            )
            w.dbfs.delete(_path, recursive=True)

    def purge(self):
        """
        Delete sink data and checkpoints
        """
        raise NotImplementedError()

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    def as_source(self, as_stream=None):
        raise NotImplementedError()

    def read(self, as_stream=None):
        """
        Read dataframe from sink.

        Parameters
        ----------
        as_stream:
            If `True`, dataframe read as stream.

        Returns
        -------
        AnyFrame
            DataFrame
        """
        return self.as_source(as_stream=as_stream).read()



from pathlib import Path


class Paths:
    def __init__(self, file):
        self.root = Path(file).parent
        self.data = self.root / "data"
        self.tmp = self.root / "tmp"

        if not self.tmp.exists():
            self.tmp.mkdir(parents=True, exist_ok=True)



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MLflowExperiment(BaseModel, PulumiResource, TerraformResource):
    """
    MLflow Experiment

    Examples
    --------
    ```py
    from laktory import models

    exp = models.resources.databricks.MLflowExperiment(
        name="/.laktory/Sample",
        artifact_location="dbfs:/tmp/my-experiment",
        description="My MLflow experiment description",
        access_controls=[
            {
                "group_name": "account users",
                "permission_level": "CAN_MANAGE",
            }
        ],
    )
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    artifact_location: str = Field(
        None,
        description="Path to dbfs:/ or s3:// artifact location of the MLflow experiment.",
    )
    creation_time: int = Field(None, description="")
    description: str = Field(
        None, description="The description of the MLflow experiment."
    )
    experiment_id: str = Field(None, description="")
    last_update_time: int = Field(None, description="")
    lifecycle_stage: str = Field(None, description="")
    name: str = Field(
        ...,
        description="""
    Name of MLflow experiment. It must be an absolute path within the Databricks workspace, e.g. 
    `/Users/<some-username>/my-experiment`. For more information about changes to experiment naming conventions,
    see [mlflow docs](https://docs.databricks.com/applications/mlflow/experiments.html#experiment-migration).
    """,
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    experiment_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MlflowExperiment"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mlflow_experiment"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AccessControlRuleSetGrantRule(BaseModel):
    """Access Control Rule Set Grant Rule"""

    role: str = Field(..., description="Role to be granted.")
    principals: list[str] = Field(
        None,
        description="a list of principals who are granted a role.",
    )


class AccessControlRuleSet(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Access Control Rule Set

    References
    ----------
    * [pulumi](https://www.pulumi.com/registry/packages/databricks/api-docs/accesscontrolruleset/)
    * [terraform](https://registry.terraform.io/providers/databricks/databricks/latest/docs/resources/access_control_rule_set)
    """

    grant_rules: list[AccessControlRuleSetGrantRule] = Field(
        None,
        description="The access control rules to be granted by this rule set, consisting of a set of principals and roles to be granted to them.",
    )
    name: str = Field(
        ...,
        description="Unique identifier of a rule set. The name determines the resource to which the rule set applies. Changing the name recreates the resource.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:AccessControlRuleSet"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "grant_rules": "grant_rules",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_access_control_rule_set"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import narwhals as nw

from laktory.datetime import utc_datetime


def current_timestamp() -> nw.Expr:
    """
    Returns the current timestamp at the start of expression evaluation as a timestamp.

    Returns
    -------
    :
        Current timestamp

    Examples
    --------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk  # noqa: F401

    df = nw.from_native(pl.DataFrame({"x": [0, 1]}))
    df = df.with_columns(tstamp=nw.laktory.current_timestamp())

    # print(df)
    ```
    """
    return nw.lit(utc_datetime())



from functools import wraps

from laktory.narwhals_ext.dataframe.display import display
from laktory.narwhals_ext.dataframe.groupby_and_agg import groupby_and_agg
from laktory.narwhals_ext.dataframe.has_column import has_column
from laktory.narwhals_ext.dataframe.schema_flat import schema_flat
from laktory.narwhals_ext.dataframe.signature import signature

# from laktory.narwhals_ext.dataframe._stream_join import stream_join
from laktory.narwhals_ext.dataframe.union import union
from laktory.narwhals_ext.dataframe.window_filter import window_filter
from laktory.narwhals_ext.dataframe.with_row_index import with_row_index
from laktory.narwhals_ext.namespace import register_anyframe_namespace
from laktory.typing import AnyFrame


@register_anyframe_namespace("laktory")
class LaktoryDataFrame:  # noqa: F811
    def __init__(self, df: AnyFrame):
        self._df = df

    @wraps(display)
    def display(self, *args, **kwargs):
        return display(self, *args, **kwargs)

    @wraps(groupby_and_agg)
    def groupby_and_agg(self, *args, **kwargs):
        return groupby_and_agg(self, *args, **kwargs)

    @wraps(has_column)
    def has_column(self, *args, **kwargs):
        return has_column(self, *args, **kwargs)

    @wraps(schema_flat)
    def schema_flat(self, *args, **kwargs):
        return schema_flat(self, *args, **kwargs)

    @wraps(signature)
    def signature(self, *args, **kwargs):
        return signature(self, *args, **kwargs)

    # @wraps(stream_join)
    # def stream_join(self, *args, **kwargs):
    #     return stream_join(self, *args, **kwargs)

    @wraps(union)
    def union(self, *args, **kwargs):
        return union(self, *args, **kwargs)

    @wraps(with_row_index)
    def with_row_index(self, *args, **kwargs):
        return with_row_index(self, *args, **kwargs)

    @wraps(window_filter)
    def window_filter(self, *args, **kwargs):
        return window_filter(self, *args, **kwargs)



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class RepoSparseCheckout(BaseModel):
    patterns: list[str] = Field(
        ...,
        description="""
    array of paths (directories) that will be used for sparse checkout. List of patterns could be updated in-place. 
    Addition or removal of the `sparse_checkout` configuration block will lead to recreation of the Git folder.
    """,
    )


class Repo(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Repo

    Examples
    --------
    ```py
    from laktory import models

    repo = models.resources.databricks.Repo(
        url="https://github.com/okube-ai/laktory",
        path="/Users/olivier.soucy@okube.ai/laktory-repo",
        branch="main",
        access_controls=[
            {"permission_level": "CAN_READ", "group_name": "account users"}
        ],
    )
    ```

    References
    ----------

    * [Databricks Repo](https://www.databricks.com/product/repos)
    * [Pulumi Databricks Repo](https://www.pulumi.com/registry/packages/databricks/api-docs/repo)

    """

    access_controls: list[AccessControl] = Field(
        [], description="List of access controls"
    )
    branch: str = Field(
        None,
        description="""
    Name of the branch for initial checkout. If not specified, the default branch of the repository will be used. 
    Conflicts with `tag`. If `branch` is removed, and `tag` isn't specified, then the repository will stay at the
    previously checked out state.
    """,
    )
    commit_hash: str = Field(
        None,
        description="""
    Hash of the HEAD commit at time of the last executed operation. It won't change if you manually perform pull 
    operation via UI or API
    """,
    )
    git_provider: str = Field(
        None,
        description="""
    Case insensitive name of the Git provider. Following values are supported right now (could be a subject for a 
    change, consult Repos API documentation): `gitHub`, `gitHubEnterprise`, `bitbucketCloud`, `bitbucketServer`, 
    `azureDevOpsServices`, `gitLab`, `gitLabEnterpriseEdition`, `awsCodeCommit`.
    """,
    )
    path: str = Field(
        None,
        description="""
    Path to put the checked out Git folder. If not specified, then the Git folder will be created in the default 
    location. If the value changes, Git folder is re-created.
    """,
    )
    sparse_checkout: list[RepoSparseCheckout] = Field(
        None, description="Sparse checkout feature in Databricks Git folders"
    )
    tag: str = Field(
        None, description="Name of the tag for initial checkout. Conflicts with branch."
    )
    url: str = Field(
        ...,
        description="The URL of the Git Repository to clone from. If the value changes, Git folder is re-created.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    repo_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Repo"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "libraries": "libraries",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_repo"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class VectorSearchIndexDeltaSyncIndexSpecEmbeddingSourceColumn(BaseModel):
    embedding_model_endpoint_name: str = Field(
        ..., description="The name of the embedding model endpoint"
    )
    name: str = Field(
        ...,
        description="Three-level name of the Vector Search Index to create (catalog.schema.index_name).",
    )


class VectorSearchIndexDeltaSyncIndexSpecEmbeddingVectorColumn(BaseModel):
    embedding_dimension: int = Field(
        ..., description="Dimension of the embedding vector."
    )
    name: str = Field(
        ...,
        description="Three-level name of the Vector Search Index to create (catalog.schema.index_name).",
    )


class VectorSearchIndexDeltaSyncIndexSpec(BaseModel):
    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "embedding_source_columns": "embedding_source_columns",
            "embedding_vector_columns": "embedding_vector_columns",
        }

    embedding_source_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingSourceColumn
    ] = Field(
        None,
        description="Array of objects representing columns that contain the embedding source",
    )
    embedding_vector_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingVectorColumn
    ] = Field(
        None,
        description="""
    (required if embedding_source_columns isn't provided) array of objects representing columns that contain the 
    embedding vectors.
    """,
    )
    embedding_writeback_table: str = Field(None, description="")
    pipeline_id: str = Field(
        None, description="ID of the associated Declarative Pipeline."
    )
    pipeline_type: Literal["TRIGGERED", "CONTINUOUS"] = Field(None, description="")
    source_table: str = Field(
        None,
        description="""
    Pipeline execution mode. Possible values are:
    - TRIGGERED: If the pipeline uses the triggered execution mode, the system stops processing after successfully 
      refreshing the source table in the pipeline once, ensuring the table is updated based on the data available when 
      the update started.
    - CONTINUOUS: If the pipeline uses continuous execution, the pipeline processes new data as it arrives in the 
      source table to keep the vector index fresh.
    """,
    )


class VectorSearchIndexDirectAccessIndexSpec(BaseModel):
    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "embedding_source_columns": "embedding_source_columns",
            "embedding_vector_columns": "embedding_vector_columns",
        }

    embedding_source_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingSourceColumn
    ] = Field(
        None,
        description="Array of objects representing columns that contain the embedding source",
    )
    embedding_vector_columns: list[
        VectorSearchIndexDeltaSyncIndexSpecEmbeddingVectorColumn
    ] = Field(
        None,
        description="""
        (required if embedding_source_columns isn't provided) array of objects representing columns that contain the
        embedding vectors.
        """,
    )
    # schema_json: str


class VectorSearchIndex(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Warehouse

    Examples
    --------
    ```py
    from laktory import models

    index = models.resources.databricks.VectorSearchIndex(
        name="dev.finances.market_news_vs_index",
        primary_key="id",
        endpoint_name="default",
        index_type="DELTA_SYNC",
        delta_sync_index_spec={
            "source_table": "dev.finances.market_news",
            "embedding_vector_columns": [
                {"name": "embedding", "embedding_dimension": 4095}
            ],
            "pipeline_type": "TRIGGERED",
        },
    )
    ```
    """

    # access_controls: list[AccessControl] = []
    endpoint_name: str = Field(
        ...,
        description="The name of the Vector Search Endpoint that will be used for indexing the data.",
    )
    index_type: Literal["DELTA_SYNC", "DIRECT_ACCESS"] = Field(
        ...,
        description="""
    Vector Search index type. Currently supported values are:
        - DELTA_SYNC: An index that automatically syncs with a source Delta Table, automatically and incrementally 
          updating the index as the underlying data in the Delta Table changes.
        - DIRECT_ACCESS: An index that supports the direct read and write of vectors and metadata through our REST and 
          SDK APIs. With this model, the user manages index updates.
    """,
    )
    primary_key: str = Field(
        ..., description="The column name that will be used as a primary key."
    )
    delta_sync_index_spec: VectorSearchIndexDeltaSyncIndexSpec = Field(
        None,
        description="Specification for Delta Sync Index. Required if index_type is DELTA_SYNC.",
    )
    direct_access_index_spec: VectorSearchIndexDirectAccessIndexSpec = Field(
        None,
        description="Specification for Direct Vector Access Index. Required if index_type is DIRECT_ACCESS.",
    )
    name: str = Field(
        ...,
        description="Three-level name of the Vector Search Index to create (catalog.schema.index_name).",
    )

    @model_validator(mode="after")
    def check_index_spec(self) -> Any:
        if self.index_type == "DELTA_SYNC" and self.delta_sync_index_spec is None:
            raise ValueError(
                "`delta_sync_index_spec` must be set with `index_type` = 'DELTA_SYNC'"
            )

        if self.index_type == "DIRECT_ACCESS" and self.direct_access_index_spec is None:
            raise ValueError(
                "`direct_access_index_spec` must be set with `index_type` = 'DIRECT_ACCESS'"
            )

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:VectorSearchIndex"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_vector_search_index"

    # @property
    # def terraform_resource_lookup_type(self) -> str:
    #     return "databricks_sql_warehouse"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Literal

from pydantic import Field

from laktory.models.basemodel import BaseModel


class FunctionGrant(BaseModel):
    """
    Privileges granted to a principal and operating on a function

    References
    ----------
    * [privilege types](https://docs.databricks.com/en/sql/language-manual/sql-ref-privileges.html#privilege-types)
    """

    principal: str = Field(..., description="User, group or service principal name")
    privileges: list[Literal["ALL_PRIVILEGES", "EXECUTE"]] = Field(
        ..., description="List of allowed privileges"
    )



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class DirectoryLookup(ResourceLookup):
    path: str = Field(
        serialization_alias="id",
        description="The absolute path of the directory, beginning with '/', e.g. '/Demo'.",
    )


class Directory(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Directory

    Examples
    --------
    ```py
    from laktory import models

    d = models.resources.databricks.Directory(path="/queries/views")
    print(d)
    '''
    resource_name_=None options=ResourceOptions(variables={}, is_enabled=True, depends_on=[], provider=None, ignore_changes=None, aliases=None, delete_before_replace=True, import_=None, parent=None, replace_on_changes=None, moved_from=None) lookup_existing=None variables={} delete_recursive=None path='/queries/views'
    '''
    print(d.resource_key)
    # > /queries/views
    print(d.resource_name)
    # > directory-queries-views
    ```
    """

    delete_recursive: Union[bool, None] = Field(
        None,
        description=" When `True`, subdirectories are also deleted when the directory is deleted",
    )
    lookup_existing: DirectoryLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    path: str = Field(
        ...,
        description="The absolute path of the directory, beginning with '/', e.g. '/pipelines'.",
    )

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        return self.path

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Directory"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_directory"



from typing import Any
from typing import Literal

import narwhals as nw
from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.basemodel import BaseModel
from laktory.models.pipelinechild import PipelineChild
from laktory.narwhals_ext.functions.sql_expr import sql_expr
from laktory.typing import AnyFrame

logger = get_logger(__name__)


class DataFrameSample(BaseModel):
    n: int = None
    fraction: float = None
    seed: int | None = None


class Watermark(BaseModel):
    """
    References
    ----------
    https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#handling-late-data-and-watermarking
    """

    column: str = Field(..., description="Event time column name")
    threshold: str = Field(
        ...,
        description="How late, expressed in seconds, the data is expected to be with respect to event time.",
    )


class BaseDataSource(BaseModel, PipelineChild):
    """
    Base class for data sources.
    """

    as_stream: bool = Field(
        False,
        description="If `True`source is read as a streaming DataFrame. Currently only supported by Spark DataFrame backend.",
    )
    # broadcast: bool = Field(
    #         False, description="If `True` DataFrame is broadcasted."
    #     )
    drop_duplicates: bool | list[str] = Field(
        None,
        description="Remove duplicated rows from source using all columns if `True` or only the provided column names.",
    )
    drops: list = Field(
        None,
        description="List of columns to drop",
    )
    filter: str = Field(
        None,
        description="SQL expression used to select specific rows from the source table",
    )
    renames: dict[str, str] = Field(
        None,
        description="Mapping between the source column names and desired column names",
    )
    # sample: DataFrameSample = None
    selects: list[str] | dict[str, str] = Field(
        None,
        description="Columns to select from the source. Can be specified as a list or as a dictionary to rename the source columns",
    )
    # watermark: Watermark | None = Field(None, description="Spark structured streaming watermark specifications")
    type: Literal["DATAFRAME", "FILE", "UNITY_CATALOG", "HIVE_METASTORE"] = Field(
        ..., description="Name of the data source type"
    )

    @model_validator(mode="after")
    def options(self) -> Any:
        with self.validate_assignment_disabled():
            if self.dataframe_backend == DataFrameBackends.PYSPARK:
                pass
            elif self.dataframe_backend == DataFrameBackends.POLARS:
                if self.as_stream:
                    raise ValueError(
                        "Streaming read is not supported with Polars Backend."
                    )
                # if self.watermark:
                #     raise ValueError("Polars DataFrames don't support watermarking.")
                # if self.broadcast:
                #     raise ValueError("Polars DataFrames don't support broadcasting.")

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        raise NotImplementedError()

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def read(self, **kwargs) -> AnyFrame:
        """
        Read data with options specified in attributes.

        Returns
        -------
        :
            Resulting dataframe
        """
        logger.info(
            f"Reading `{self.__class__.__name__}` {self._id} with {self.dataframe_backend}"
        )
        df = self._read(**kwargs)

        # Convert to Narwhals
        if not isinstance(df, (nw.LazyFrame, nw.DataFrame)):
            df = nw.from_native(df)

        # Post read
        df = self._post_read(df)

        logger.info("Read completed.")

        return df

    def _read(self, **kwargs) -> AnyFrame:
        if self.dataframe_backend == DataFrameBackends.PYSPARK:
            return self._read_spark()

        if self.dataframe_backend == DataFrameBackends.POLARS:
            return self._read_polars(**kwargs)

        raise ValueError(f"`{self.dataframe_backend}` not supported for `{type(self)}`")

    def _read_spark(self, **kwargs) -> nw.LazyFrame:
        raise NotImplementedError(
            f"`{self.dataframe_backend}` not supported for `{type(self)}`"
        )

    def _read_polars(self, **kwargs) -> nw.LazyFrame:
        raise NotImplementedError(
            f"`{self.dataframe_backend}` not supported for `{type(self)}`"
        )

    def _post_read(self, df: AnyFrame) -> AnyFrame:
        # Apply filter
        if self.filter:
            df = df.filter(sql_expr(self.filter))

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [sql_expr(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [sql_expr(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Renames
        if self.renames:
            df = df.rename(self.renames)

        # Apply drops
        if self.drops:
            df = df.drop(*self.drops, strict=False)

        # # Apply Watermark
        # if self.watermark:
        #     df = df.withWatermark(
        #         self.watermark.column,
        #         self.watermark.threshold,
        #     )
        #
        # # Broadcast
        # if self.broadcast:
        #     df = F.broadcast(df)

        # Drop Duplicates
        if self.drop_duplicates:
            subset = None
            if isinstance(self.drop_duplicates, list):
                subset = self.drop_duplicates

            df = df.unique(subset=subset)

        # Sample
        # TODO: Enable when Narwhals support sampling on LazyFrame
        # if self.sample:
        #     df = df.sample(
        #         n=self.sample.n, fraction=self.sample.fraction, seed=self.sample.seed
        #     )

        return df



from typing import Any
from typing import Literal
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class PipelineTriggerCron(BaseModel):
    quartz_cron_schedule: str = Field(None, description="")
    timezone_id: str = Field(None, description="")


class PipelineTrigger(BaseModel):
    cron: PipelineTriggerCron = Field(None, description="")
    # manual:


class PipelineRunAs(BaseModel):
    service_principal_name: str = Field(None, description="")
    user_name: str = Field(None, description="")


class PipelineRestartWindow(BaseModel):
    start_hour: int = Field(..., description="")
    days_of_weeks: list[str] = Field(None, description="")
    time_zone_id: str = Field(None, description="")


class PipelineLatestUpdate(BaseModel):
    creation_time: str = Field(None, description="")
    state: str = Field(None, description="")
    update_id: str = Field(None, description="")


class PipelineGatewayDefinition(BaseModel):
    connection_id: str = Field(
        None,
        description="Immutable. The Unity Catalog connection this gateway pipeline uses to communicate with the source.",
    )
    connection_name: str = Field(..., description="")
    gateway_storage_catalog: str = Field(
        ...,
        description="Required, Immutable. The name of the catalog for the gateway pipeline's storage location.",
    )
    gateway_storage_name: str = Field(
        None,
        description="""
        Required. The Unity Catalog-compatible naming for the gateway storage location. This is the destination to use 
        for the data that is extracted by the gateway. Delta Live Tables system will automatically create the storage 
        location under the catalog and schema.
        """,
    )
    gateway_storage_schema: str = Field(
        ...,
        description="Required, Immutable. The name of the schema for the gateway pipelines's storage location.",
    )


class PipelineDeployment(BaseModel):
    kind: str = Field(
        ..., description="The deployment method that manages the pipeline."
    )
    metadata_file_path: str = Field(
        ...,
        description="The path to the file containing metadata about the deployment.",
    )


class PipelineEventLog(BaseModel):
    name: str = Field(
        ..., description="The table name the event log is published to in UC."
    )
    catalog: str = Field(
        ..., description="The UC catalog the event log is published under."
    )
    schema_: str = Field(
        ...,
        description="The UC schema the event log is published under.",
        validation_alias=AliasChoices("schema", "schema_"),
    )


class PipelineFilters(BaseModel):
    excludes: str = Field(..., description="Paths to exclude.")
    includes: str = Field(..., description="Paths to include.")


class PipelineLibraryFile(BaseModel):
    path: str = Field(..., description="")


class PipelineLibraryNotebook(BaseModel):
    path: str = Field(..., description="Workspace notebook filepath")


class PipelineLibrary(BaseModel):
    file: str = Field(None, description="File specifications")
    notebook: PipelineLibraryNotebook = Field(
        None, description="Notebook specifications"
    )


class PipelineNotifications(BaseModel):
    alerts: list[
        Literal[
            "on-update-success",
            "on-update-failure",
            "on-update-fatal-failure",
            "on-flow-failure",
        ]
    ] = Field(..., description="Alert types")
    recipients: list[str] = Field(
        ..., description="List of user/group/service principal names"
    )


class PipelineCluster(Cluster):
    """
    Pipeline Cluster. Same attributes as `laktory.models.Cluster`, except for

    * `autotermination_minutes`
    * `cluster_id`
    * `data_security_mode`
    * `enable_elastic_disk`
    * `idempotency_token`
    * `is_pinned`
    * `libraries`
    * `no_wait`
    * `node_type_id`
    * `runtime_engine`
    * `single_user_name`
    * `spark_version`

    that are not allowed.
    """

    autotermination_minutes: int = Field(None, exclude=True)
    cluster_id: str = Field(None, exclude=True)
    data_security_mode: str = Field(None, exclude=True)
    enable_elastic_disk: bool = Field(None, exclude=True)
    idempotency_token: str = Field(None, exclude=True)
    is_pinned: bool = Field(None, exclude=True)
    libraries: list[Any] = Field(None, exclude=True)
    node_type_id: str = None
    no_wait: bool = Field(None, exclude=True)
    runtime_engine: str = Field(None, exclude=True)
    single_user_name: str = Field(None, exclude=True)
    spark_version: str = Field(None, exclude=True)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "autotermination_minutes",
            "cluster_id",
            "data_security_mode",
            "enable_elastic_disk",
            "idempotency_token",
            "is_pinned",
            "libraries",
            "no_wait",
            "runtime_engine",
            "single_user_name",
            "spark_version",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self


class Pipeline(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Lakeflow Declarative Pipeline (formerly Delta Live Tables)

    Examples
    --------
    Assuming the configuration yaml file
    ```py
    import io

    from laktory import models

    # Define pipeline
    pipeline_yaml = '''
    name: pl-stock-prices

    catalog: dev
    target: finance

    clusters:
      - name : default
        node_type_id: Standard_DS3_v2
        autoscale:
          min_workers: 1
          max_workers: 2

    libraries:
      - notebook:
          path: /pipelines/dlt_brz_template.py
      - notebook:
          path: /pipelines/dlt_slv_template.py
      - notebook:
          path: /pipelines/dlt_gld_stock_performances.py

    access_controls:
      - group_name: account users
        permission_level: CAN_VIEW
      - group_name: role-engineers
        permission_level: CAN_RUN

    '''
    pipeline = models.resources.databricks.Pipeline.model_validate_yaml(
        io.StringIO(pipeline_yaml)
    )
    ```

    References
    ----------

    * [Databricks Pipeline](https://docs.databricks.com/api/workspace/pipelines/create)
    * [Pulumi Databricks Pipeline](https://www.pulumi.com/registry/packages/databricks/api-docs/pipeline/)
    """

    access_controls: list[AccessControl] = Field(
        [], description="Pipeline access controls"
    )
    allow_duplicate_names: bool = Field(
        None,
        description="""
        If `False`, deployment will fail if name conflicts with that of another pipeline.
        """,
    )
    budget_policy_id: str = Field(
        None,
        description="optional string specifying ID of the budget policy for this DLT pipeline.",
    )
    catalog: Union[str, None] = Field(
        None, description="Name of the unity catalog storing the pipeline tables"
    )
    cause: str = Field(None, description="")
    channel: Literal["CURRENT", "PREVIEW"] = Field(
        "PREVIEW",
        description="Name of the release channel for Spark version used by DLT pipeline.",
    )
    cluster_id: str = Field(None, description="")
    clusters: list[PipelineCluster] = Field(
        [],
        description="""
        Clusters to run the pipeline. If none is specified, pipelines will automatically select a default cluster 
        configuration for the pipeline.
        """,
    )
    creator_user_name: str = Field(None, description="")
    configuration: dict[str, str] = Field(
        {},
        description="List of values to apply to the entire pipeline. Elements must be formatted as key:value pairs",
    )
    continuous: bool = Field(
        None, description="If `True`, the pipeline is run continuously."
    )
    deployment: PipelineDeployment = Field(
        None, description="Deployment type of this pipeline."
    )
    development: bool = Field(
        None, description="If `True` the pipeline is run in development mode"
    )
    edition: Literal["CORE", "PRO", "ADVANCED"] = Field(
        None, description="Name of the product edition"
    )
    event_log: PipelineEventLog = Field(
        None,
        description="An optional block specifying a table where DLT Event Log will be stored.",
    )
    expected_last_modified: int = Field(None, description="")
    filters: PipelineFilters = Field(
        None,
        description="Filters on which Pipeline packages to include in the deployed graph.",
    )
    gateway_definition: PipelineGatewayDefinition = Field(
        None, description="The definition of a gateway pipeline to support CDC."
    )
    health: str = Field(None, description="")
    # ingestion_definition: = Field(None, description="")  #TODO
    last_modified: int = Field(None, description="")
    latest_updates: list[PipelineLatestUpdate] = Field(None, description="")
    libraries: list[PipelineLibrary] = Field(
        None, description="Specifies pipeline code (notebooks) and required artifacts."
    )
    name: str = Field(..., description="Pipeline name")
    name_prefix: str = Field(None, description="Prefix added to the DLT pipeline name")
    name_suffix: str = Field(None, description="Suffix added to the DLT pipeline name")
    notifications: list[PipelineNotifications] = Field(
        [], description="Notifications specifications"
    )
    photon: bool = Field(None, description="If `True`, Photon engine enabled.")
    restart_window: PipelineRestartWindow = Field(None, description="")
    root_path: str = Field(
        None,
        description="""
        An optional string specifying the root path for this pipeline. This is used as the root directory when editing
        the pipeline in the Databricks user interface and it is added to sys.path when executing Python sources during 
        pipeline execution.
        """,
    )
    run_as: PipelineRunAs = Field(None, description="")
    run_as_user_name: str = Field(None, description="")
    schema_: str = Field(
        None,
        description="""
        The default schema (database) where tables are read from or published to. The presence of this
        attribute implies that the pipeline is in direct publishing mode.
        """,
        validation_alias=AliasChoices("schema", "schema_"),
    )
    serverless: bool = Field(None, description="If `True`, serverless is enabled")
    state: str = Field(None, description="")
    storage: str = Field(
        None,
        description="""
        A location on DBFS or cloud storage where output data and metadata required for pipeline execution are stored.
        By default, tables are stored in a subdirectory of this location. Change of this parameter forces recreation of
        the pipeline. (Conflicts with `catalog`).
        """,
    )
    tags: dict[str, str] = Field(
        None,
        description="""
        A map of tags associated with the pipeline. These are forwarded to the cluster as cluster tags, and are 
        therefore subject to the same limitations. A maximum of 25 tags can be added to the pipeline.
        """,
    )
    target: str = Field(
        None,
        description="""
        The name of a database (in either the Hive metastore or in a UC catalog) for persisting pipeline output data.
        Configuring the target setting allows you to view and query the pipeline output data from the Databricks UI.
        """,
    )
    trigger: PipelineTrigger = Field(None, description="")
    url: str = Field(
        None, description="URL of the DLT pipeline on the given workspace."
    )

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.name = self.name_prefix + self.name
                self.name_prefix = ""
            if self.name_suffix:
                self.name = self.name + self.name_suffix
                self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        """
        dlt
        """
        return "dlt-pipeline"

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []

        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    pipeline_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_renames(self) -> dict[str, str]:
        return {"schema_": "schema"}

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Pipeline"

    @property
    def pulumi_excludes(self) -> list[str] | dict[str, bool]:
        return {
            "access_controls": True,
            "clusters": {"__all__": {"access_controls"}},
            "name_prefix": True,
            "name_suffix": True,
        }

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        k = "clusters"
        if k in d:
            _clusters = []
            for c in d[k]:
                c["label"] = c.pop("clusterName")
                _clusters += [c]
            d[k] = _clusters
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_pipeline"

    @property
    def terraform_renames(self) -> dict[str, str]:
        return self.pulumi_renames

    @property
    def terraform_excludes(self) -> list[str] | dict[str, bool]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        k = "cluster"
        if k in d:
            _clusters = []
            for c in d[k]:
                c["label"] = c.pop("cluster_name")
                _clusters += [c]
            d[k] = _clusters
        return d



from datetime import datetime

import narwhals as nw

import laktory as lk


@lk.api.register_anyframe_namespace("lake")
class LakeNamespace:
    def __init__(self, _df):
        self._df = _df

    def with_last_modified(self):
        return self._df.with_columns(last_modified=nw.lit(datetime.now()))



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.groupmember import GroupMember
from laktory.models.resources.databricks.mwspermissionassignment import (
    MwsPermissionAssignment,
)
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class GroupLookup(ResourceLookup):
    id: str = Field(
        None, description="Id of the group. Only supported when using Pulumi backend."
    )
    display_name: str = Field(
        None,
        description="Display name of the group. Only support when using Terraform backend",
    )


class Group(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks group

    Examples
    --------
    ```py
    from laktory import models

    d = models.resources.databricks.Group(display_name="role-engineers")
    ```
    """

    acl_principal_id: str = Field(
        None,
        description="Identifier for use in databricks_access_control_rule_set, e.g. `groups/Some Group`.",
    )
    allow_cluster_create: bool = Field(
        False,
        description="""
    This is a field to allow the group to have cluster create privileges. More fine grained permissions could be 
    assigned with databricks.Permissions and cluster_id argument. Everyone without `allow_cluster_create` argument set,
    but with permission to use Cluster Policy would be able to create clusters, but within boundaries of that specific 
    policy.
    """,
    )
    allow_instance_pool_create: bool = Field(
        None,
        description="""
    This is a field to allow the group to have instance pool create privileges. More fine grained permissions could 
    be assigned with databricks.Permissions and instance_pool_id argument.
    """,
    )
    databricks_sql_access: bool = Field(
        None,
        description="""
    This is a field to allow the group to have access to Databricks SQL feature in User Interface and through 
    databricks_sql_endpoint.
    """,
    )
    display_name: str = Field(None, description="Display name for the group.")
    external_id: str = Field(
        None, description="ID of the group in an external identity provider."
    )
    force: bool = Field(
        None,
        description="""
    Ignore `cannot create group: Group with name X already exists.` errors and implicitly import the specific group 
    into IaC state, enforcing entitlements defined in the instance of resource. This functionality is experimental and 
    is designed to simplify corner cases, like Azure Active Directory synchronisation.
    """,
    )
    lookup_existing: GroupLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    member_ids: list[str] = Field(
        [],
        description="A list of all member ids of the group. Can be users, groups or service principals",
    )
    url: str = Field(None, description="")
    workspace_access: bool = Field(
        None, description="When `True`, the group is allowed to have workspace access"
    )
    workspace_permission_assignments: list[MwsPermissionAssignment] = Field(
        None, description="Workspace access privileges"
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """display name or id"""
        lookup_display_name = (
            self.lookup_existing.display_name if self.lookup_existing else None
        )
        lookup_id = self.lookup_existing.id if self.lookup_existing else None
        return next(
            (
                attr
                for attr in (
                    self.display_name,
                    self.external_id,
                    lookup_display_name,
                    lookup_id,
                )
                if attr is not None
            ),
            "",
        )

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Group"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["member_ids", "workspace_permission_assignments"]

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace permission assignments
        """
        resources = []

        # Group Members
        for member_id in self.member_ids:
            resources += [
                GroupMember(
                    group_id=f"${{resources.{self.resource_name}.id}}",
                    member_id=member_id,
                )
            ]

        if self.workspace_permission_assignments:
            for a in self.workspace_permission_assignments:
                if a.principal_id is None:
                    a.principal_id = f"${{resources.{self.resource_name}.id}}"
                resources += [a]
        return resources

    # TODO:
    # if group.id is None:
    #     self.group = databricks.Group(name, opts=opts, **group.model_pulumi_dump())
    #     group.id = self.group.id
    # else:
    #     self.group = databricks.Group.get(name, id=group.id)

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_group"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import narwhals as nw


class NameSpace:
    def __init__(self, name: str, ns_cls: type):
        self._name = name
        self._ns_cls = ns_cls

    def __get__(self, instance, owner):
        if instance is None:
            return self._ns_cls
        ns_instance = self._ns_cls(instance)
        setattr(instance, self._name, ns_instance)
        return ns_instance


def register_expr_namespace(name: str):
    """
    Decorator for registering custom method to a Narwhals Expression.

    Parameters
    ----------
    name:
        Name of the namespace

    Examples
    -------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk


    @lk.api.register_expr_namespace("custom")
    class CustomNamespace:
        def __init__(self, _expr):
            self._expr = _expr

        def double(self):
            return self._expr * 2


    df = nw.from_native(pl.DataFrame({"x": [0, 1]}))

    df = df.with_columns(x2=nw.col("x").custom.double())

    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |    | x | x2 |    |
    |    |---|----|    |
    |    | 0 | 0  |    |
    |    | 1 | 2  |    |
    └──────────────────┘
    '''
    ```

    References
    ----------
    * [Narwhals Extension](https://www.laktory.ai/concepts/extension_custom/)
    """

    def wrapper(ns_cls: type):
        setattr(nw.Expr, name, NameSpace(name, ns_cls))
        return ns_cls

    return wrapper


def register_anyframe_namespace(name: str):
    """
    Decorator for registering custom method to a Narwhals DataFrame and LazyFrame.

    Parameters
    ----------
    name:
        Name of the namespace

    Examples
    -------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk


    @lk.api.register_anyframe_namespace("custom")
    class CustomNamespace:
        def __init__(self, _df):
            self._df = _df

        def with_x2(self):
            return self._df.with_columns(x2=nw.col("x") * 2)


    df = nw.from_native(pl.DataFrame({"x": [0, 1]}))

    df = df.custom.with_x2()

    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |    | x | x2 |    |
    |    |---|----|    |
    |    | 0 | 0  |    |
    |    | 1 | 2  |    |
    └──────────────────┘
    '''
    ```

    References
    ----------
    * [Narwhals Extension](https://www.laktory.ai/concepts/extension_custom/)
    """

    def wrapper(ns_cls: type):
        setattr(nw.DataFrame, name, NameSpace(name, ns_cls))
        setattr(nw.LazyFrame, name, NameSpace(name, ns_cls))
        return ns_cls

    return wrapper


def register_dataframe_namespace(name: str):
    """
    Decorator for registering custom method to a Narwhals DataFrame.

    Parameters
    ----------
    name:
        Name of the namespace

    Examples
    -------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk


    @lk.api.register_dataframe_namespace("custom")
    class CustomNamespace:
        def __init__(self, _df):
            self._df = _df

        def with_x2(self):
            return self._df.with_columns(x2=nw.col("x") * 2)


    df = nw.from_native(pl.DataFrame({"x": [0, 1]}))

    df = df.custom.with_x2()

    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |    | x | x2 |    |
    |    |---|----|    |
    |    | 0 | 0  |    |
    |    | 1 | 2  |    |
    └──────────────────┘
    '''
    ```

    References
    ----------
    * [Narwhals Extension](https://www.laktory.ai/concepts/extension_custom/)
    """

    def wrapper(ns_cls: type):
        setattr(nw.DataFrame, name, NameSpace(name, ns_cls))
        return ns_cls

    return wrapper


def register_lazyframe_namespace(name: str):
    """
    Decorator for registering custom method to a Narwhals LazyFrame.

    Parameters
    ----------
    name:
        Name of the namespace

    Examples
    -------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk


    @lk.api.register_lazyframe_namespace("custom")
    class CustomNamespace:
        def __init__(self, _df):
            self._df = _df

        def with_x2(self):
            return self._df.with_columns(x2=nw.col("x") * 2)


    df = nw.from_native(pl.DataFrame({"x": [0, 1]}).lazy())

    df = df.custom.with_x2()

    print(df.collect().sort("x"))
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |    | x | x2 |    |
    |    |---|----|    |
    |    | 0 | 0  |    |
    |    | 1 | 2  |    |
    └──────────────────┘
    '''
    ```

    References
    ----------
    * [Narwhals Extension](https://www.laktory.ai/concepts/extension_custom/)
    """

    def wrapper(ns_cls: type):
        setattr(nw.LazyFrame, name, NameSpace(name, ns_cls))
        return ns_cls

    return wrapper



import re

from pydantic import Field

from laktory.models.basemodel import BaseModel


class BaseProvider(BaseModel):
    alias: str = Field(None, exclude=False)
    source: str = Field(None, exclude=True)
    version: str = Field(None, exclude=True)

    @property
    def resource_name(self) -> str | None:
        """
        dots (.) are allowed to support terraform aliases
        """

        # if self.resource_name_:
        name = self.resource_name_
        # else:
        #     name = self.default_resource_name

        if name is None:
            return None

        alias_pattern = ""
        if self.alias is not None:
            alias_pattern = "."
        pattern = re.compile(r"^[a-zA-Z][a-zA-Z0-9-_" + alias_pattern + "]*$")

        if self.alias:
            if not name.endswith(f".{self.alias}"):
                raise ValueError(
                    f"Resource name `{name}` is invalid. A name for alias provider must end with .{self.alias}"
                )

        if not pattern.match(name):
            raise ValueError(
                f"Resource name `{name}` is invalid. A name must start with a letter or underscore and may contain only letters, digits, underscores, dots (if alias), and dashes."
            )

        return name

    @property
    def resource_name_without_alias(self):
        if self.alias is None:
            return self.resource_name
        return self.resource_name.replace(f".{self.alias}", "")

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self):
        return ["source", "version"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return None



from .orchestrators.databricksjoborchestrator import DatabricksJobOrchestrator
from .orchestrators.databrickspipelineorchestrator import DatabricksPipelineOrchestrator
from .pipeline import Pipeline
from .pipeline import PipelineNode



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.grants.metastoregrant import MetastoreGrant
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.metastoreassignment import MetastoreAssignment
from laktory.models.resources.databricks.metastoredataaccess import MetastoreDataAccess
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MetastoreLookup(ResourceLookup):
    metastore_id: str = Field(
        serialization_alias="id", description="ID of the metastore"
    )


class Metastore(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Metastore

    Examples
    --------
    ```py
    ```
    """

    cloud: str = Field(None, description="")
    created_at: int = Field(None, description="")
    created_by: str = Field(None, description="")
    data_accesses: list[MetastoreDataAccess] = Field(
        None, description="List of data accesses (storage credentials)"
    )
    default_data_access_config_id: str = Field(None, description="")
    delta_sharing_organization_name: str = Field(
        None,
        description="""
    The organization name of a Delta Sharing entity. This field is used for Databricks to Databricks sharing. Once 
    this is set it cannot be removed and can only be modified to another valid value. To delete this value please 
    taint and recreate the resource.
    """,
    )
    delta_sharing_recipient_token_lifetime_in_seconds: int = Field(
        None,
        description="""
    Required along with `delta_sharing_scope`. Used to set expiration duration in seconds on recipient data access 
    tokens. Set to 0 for unlimited duration.
    """,
    )
    delta_sharing_scope: str = Field(
        None,
        description="""
    Required along with delta_sharing_recipient_token_lifetime_in_seconds. Used to enable delta sharing on the 
    metastore. Valid values: INTERNAL, INTERNAL_AND_EXTERNAL.
    """,
    )
    force_destroy: bool = Field(
        None, description="Destroy metastore regardless of its contents."
    )
    global_metastore_id: str = Field(None, description="")
    grant: Union[MetastoreGrant, list[MetastoreGrant]] = Field(
        None,
        description="""
    Grant(s) operating on the Metastore and authoritative for a specific principal. Other principals within the grants 
    are preserved. Mutually exclusive with `grants`.
    """,
    )
    grants: list[MetastoreGrant] = Field(
        None,
        description="""
    Grants operating on the Metastore and authoritative for all principals. Replaces any existing grants defined inside
    or outside of Laktory. Mutually exclusive with `grant`.
    """,
    )
    grants_provider: str = Field(None, description="Provider used for deploying grants")
    lookup_existing: MetastoreLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    metastore_id: str = Field(None, description="")
    name: str = Field(None, description="Name of metastore.")
    owner: str = Field(
        None, description="Username/groupname/sp application_id of the metastore owner."
    )
    region: str = Field(None, description="The region of the metastore")
    storage_root: str = Field(
        None,
        description="""
    Path on cloud storage account, where managed databricks.Table are stored. Change forces creation of a new resource.
     If no storage_root is defined for the metastore, each catalog must have a storage_root defined.
    """,
    )
    storage_root_credential_id: Union[str, None] = Field(None, description="")
    updated_at: int = Field(None, description="")
    updated_by: str = Field(None, description="")
    workspace_assignments: list[MetastoreAssignment] = Field(
        None, description="List of workspace to which metastore is assigned to"
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """
        Resource key used to build default resource name. Equivalent to
        name properties if available. Otherwise, empty string.
        """
        key = self.name
        if key is None and self.lookup_existing:
            key = self.lookup_existing.metastore_id
        return key

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace assignments
        - grants
        """
        resources = []

        depends_on = []
        if self.workspace_assignments:
            for a in self.workspace_assignments:
                a.metastore_id = f"${{resources.{self.resource_name}.id}}"
                depends_on += [f"${{resources.{a.resource_name}}}"]
                resources += [a]

        options = {"provider": self.grants_provider}
        if depends_on:
            options["depends_on"] = depends_on

        _resources = self.get_grants_additional_resources(
            object={"metastore": f"${{resources.{self.resource_name}.id}}"},
            options=options,
        )
        for r in _resources:
            depends_on += [f"${{resources.{r.resource_name}}}"]
        resources += _resources

        if self.data_accesses:
            for data_access in self.data_accesses:
                data_access.metastore_id = f"${{resources.{self.resource_name}.id}}"
                _core_resources = data_access.core_resources
                for r in _core_resources[1:]:
                    if r.options.provider is None:
                        r.options.provider = self.grants_provider
                    if depends_on:
                        r.options.depends_on = depends_on
                resources += _core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Metastore"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "workspace_assignments",
            "grant",
            "grants",
            "grants_provider",
            "data_accesses",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_metastore"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes + ["cloud"]



QUICKSTART_TEMPLATES = [
    "unity-catalog",
    "workspace",
    "workflows",
    "local-pipeline",
]

# Convert to enums?
SUPPORTED_BACKENDS = [
    "pulumi",
    "terraform",
]


# CACHE_ROOT = "./laktory"
CACHE_ROOT = "./"
